import os
import sys
import numpy as np
from random import randint
import pyqtgraph as pg
from PyQt5.QtWidgets import *
from PyQt5.QtGui import *
from PyQt5.QtCore import *
from .wtiles import QDoubleButton
from .uuuuuu import read_qss_file


eye_button_style = """
    border-left: None;
    border-right: 1px solid gray;
    border-top: None;
    border-bottom: None;
    border-radius: 0px;
    background: transparent;
"""


line_edit_style = """
QLineEdit {
    background: transparent;
    border: 1px solid green;
    border-radius:0px;
    text-align:left; 
    color: white; 
    padding-left: 5px;    
    margin: 0px;
    height: 40px;
    width: 220px; 
}
"""


class ObjectTextButton(QPushButton):
    def __init__(self):
        QPushButton.__init__(self)
        btn_style = read_qss_file("qss/object_text_button.qss")
        self.setStyleSheet(btn_style)
        self.setFixedSize(QSize(150, 40))


class CompareWindow(QDialog):
    def __init__(self, obj_names, obj_data):
        super().__init__()

        self.setWindowTitle("Compare Information Window")

        n_object = len(obj_names)
        # print(n_object)
        # print(obj_data)
        all_label_names = []
        all_label_acronym = []
        all_label_color = []
        all_label_channels = []
        for i in range(n_object):
            all_label_names.append(obj_data[i]["label_name"])
            all_label_acronym.append(obj_data[i]["label_acronym"])
            all_label_color.append(obj_data[i]["label_color"])
            all_label_channels.append(obj_data[i]["region_sites"])
        all_label_names = np.concatenate(all_label_names)
        all_label_acronym = np.concatenate(all_label_acronym)
        all_label_color = np.concatenate(all_label_color)
        all_label_channels = np.concatenate(all_label_channels)

        # print(all_label_names)
        # print(all_label_acronym)
        # print(all_label_color)
        # print(all_label_channels)

        unique_label_names = np.unique(all_label_names)
        unique_label_acronym = []
        unique_label_color = []
        unique_label_channels = []
        for i in range(len(unique_label_names)):
            c_ind = np.where(all_label_names == unique_label_names[i])[0]
            unique_label_acronym.append(all_label_acronym[c_ind[0]])
            unique_label_color.append(all_label_color[c_ind[0]])
            unique_label_channels.append(np.sum(all_label_channels[c_ind]))

        layout = QVBoxLayout()
        full_name = obj_names[0]
        for i in range(1, n_object):
            full_name = full_name + ", "
            full_name = full_name + obj_names[i]

        self.label = QLabel("Compare {}".format(full_name))
        color = QColor(128, 128, 128, 128)
        label_style = "QLabel {background-color: " + color.name() + "; font-size: 20px}"
        self.label.setStyleSheet(label_style)

        sec_group = QGroupBox()
        slayout = QGridLayout(sec_group)
        lb1 = QLabel("Brain Region")
        lb2 = QLabel("Acronym")
        lb3 = QLabel("Channels")
        lb4 = QLabel("Color")
        slayout.addWidget(lb1, 0, 0, 1, 1)
        slayout.addWidget(lb2, 0, 1, 1, 1)
        slayout.addWidget(lb3, 0, 2, 1, 1)
        slayout.addWidget(lb4, 0, 3, 1, 1)

        channels = np.ravel(unique_label_channels).astype(int)

        for i in range(len(unique_label_names)):
            slayout.addWidget(QLabel(unique_label_names[i]), i + 1, 0, 1, 1)
            slayout.addWidget(QLabel(unique_label_acronym[i]), i + 1, 1, 1, 1)
            slayout.addWidget(QLabel(str(channels[i])), i + 1, 2, 1, 1)
            clb = QLabel()
            da_color = QColor(
                unique_label_color[i][0],
                unique_label_color[i][1],
                unique_label_color[i][2],
                255,
            ).name()
            clb.setStyleSheet(
                "QLabel {background-color: " + da_color + "; width: 20px; height: 20px}"
            )
            slayout.addWidget(clb, i + 1, 3, 1, 1)

        # make plot data
        plot_frame = QFrame()
        plot_frame.setMaximumWidth(300)
        plot_frame.setMinimumWidth(300)
        view_layout = QHBoxLayout(plot_frame)
        view_layout.setSpacing(0)
        view_layout.setContentsMargins(0, 0, 0, 0)

        w = pg.GraphicsLayoutWidget()
        view = w.addViewBox()
        view.setAspectLocked()

        view_layout.addWidget(w)

        for k in range(n_object):
            vis_data = obj_data[k]["vis_data"]
            n_column = len(vis_data)

            plot_center_x_val = k * 10 + 0.5 * n_column

            probe_type_name = obj_data[k]["probe_type_name"]
            # draw tips
            if probe_type_name != "Tetrode":
                da_tip_loc = np.array(
                    [[0, 0], [0.5 * n_column, -2 * n_column], [n_column, 0]]
                )
                da_tip_loc = da_tip_loc + np.array([plot_center_x_val, 0])
                tips = pg.PlotDataItem(
                    da_tip_loc,
                    connect="all",
                    fillLevel=0,
                    fillBrush=pg.mkBrush(color=(128, 128, 128)),
                )
                view.addItem(tips)

            # draw area
            region_label = obj_data[k]["region_label"]

            group_color = obj_data[k]["label_color"]
            sites_label = obj_data[k]["sites_label"]
            for i in range(n_column):
                group_id = vis_data[i]["group_id"].astype(int)
                start_loc = vis_data[i]["start_loc"]
                end_loc = vis_data[i]["end_loc"]
                sites_loc_column = vis_data[i]["sites"]

                for j in range(len(group_id)):
                    area_data = np.array([[i, end_loc[j]], [i + 1, end_loc[j]]])
                    area_data = area_data + np.array([plot_center_x_val, 0])
                    da_item = pg.PlotDataItem(
                        area_data,
                        fillLevel=start_loc[j],
                        fillBrush=group_color[group_id[j]],
                        pen=None,
                    )
                    view.addItem(da_item)

                sites_color = []
                for j in range(len(sites_label[i])):
                    color_ind = np.where(region_label == sites_label[i][j])[0][0]
                    sites_color.append(group_color[color_ind])

                pnt_data = np.stack(
                    [np.repeat(i + 0.5, len(sites_loc_column)), sites_loc_column],
                    axis=1,
                )
                pnt_data = pnt_data + np.array([plot_center_x_val, 0])
                da_item = pg.ScatterPlotItem(
                    pos=pnt_data,
                    pen=(128, 128, 128, 128),
                    brush=sites_color,
                    symbol="s",
                    size=3,
                    hoverSize=8,
                )
                view.addItem(da_item)

        channel_info_frame = QFrame()
        channel_info_layout = QHBoxLayout(channel_info_frame)
        channel_info_layout.setContentsMargins(0, 0, 0, 0)
        channel_info_layout.setSpacing(10)
        channel_info_layout.addWidget(plot_frame)
        channel_info_layout.addWidget(sec_group)

        # ok button, used to close window
        ok_btn = QDialogButtonBox(QDialogButtonBox.Ok)
        ok_btn.accepted.connect(self.accept)

        # add widget to layout
        layout.addWidget(self.label)
        layout.addWidget(channel_info_frame)
        layout.addWidget(ok_btn)
        self.setLayout(layout)

    def accept(self) -> None:
        self.close()


class CellsInfoWindow(QDialog):
    def __init__(self, name, data):
        super().__init__()

        self.setWindowTitle("Cell Information Window")

        layout = QVBoxLayout()
        self.label = QLabel(name)
        color = QColor(
            data["vis_color"][0],
            data["vis_color"][1],
            data["vis_color"][2],
            data["vis_color"][3],
        )
        label_style = "QLabel {background-color: " + color.name() + "; font-size: 20px}"
        self.label.setStyleSheet(label_style)

        data_mat = data["data"][0]
        for i in range(1, len(data["data"])):
            data_mat = np.vstack([data_mat, data["data"][i]])

        sec_group = QGroupBox("Total Count: {}".format(len(data_mat)))
        slayout = QGridLayout(sec_group)
        lb1 = QLabel("Brain Region")
        lb2 = QLabel("Acronym")
        lb3 = QLabel("Color")
        lb4 = QLabel("Count")
        slayout.addWidget(lb1, 0, 0, 1, 1)
        slayout.addWidget(lb2, 0, 1, 1, 1)
        slayout.addWidget(lb3, 0, 2, 1, 1)
        slayout.addWidget(lb4, 0, 3, 1, 1)

        for i in range(len(data["label_name"])):
            slayout.addWidget(QLabel(data["label_name"][i]), i + 1, 0, 1, 1)
            slayout.addWidget(QLabel(data["label_acronym"][i]), i + 1, 1, 1, 1)
            clb = QLabel()
            da_color = QColor(
                data["label_color"][i][0],
                data["label_color"][i][1],
                data["label_color"][i][2],
                255,
            ).name()
            clb.setStyleSheet(
                "QLabel {background-color: " + da_color + "; width: 20px; height: 20px}"
            )
            slayout.addWidget(clb, i + 1, 2, 1, 1)
            slayout.addWidget(QLabel(str(data["region_count"][i])), i + 1, 3, 1, 1)

        # ok button, used to close window
        ok_btn = QDialogButtonBox(QDialogButtonBox.Ok)
        ok_btn.accepted.connect(self.accept)

        # add widget to layout
        layout.addWidget(self.label)
        layout.addWidget(sec_group)
        layout.addWidget(ok_btn)
        self.setLayout(layout)

    def accept(self) -> None:
        self.close()

    def set_probe_color(self, color):
        self.label.setStyleSheet("QLabel {background-color: " + color + ";}")


class VirusInfoWindow(QDialog):
    def __init__(self, name, data):
        super().__init__()

        self.setWindowTitle("Virus Information Window")

        layout = QVBoxLayout()
        self.label = QLabel(name)
        color = QColor(
            data["vis_color"][0],
            data["vis_color"][1],
            data["vis_color"][2],
            data["vis_color"][3],
        )
        label_style = "QLabel {background-color: " + color.name() + "; font-size: 20px}"
        self.label.setStyleSheet(label_style)

        region_label = data["label_id"]
        region_name = data["label_name"]
        region_acronym = data["label_acronym"]
        virus_volume = data["virus_volume"]
        region_volume = data["region_volume"]
        region_color = data["label_color"]
        proportion = np.round(np.ravel(virus_volume) / np.ravel(region_volume) * 100, 2)

        sec_group = QGroupBox()
        sec_layout = QGridLayout(sec_group)
        column_names = [
            QLabel("ID"),
            QLabel("Brain Region"),
            QLabel("Acronym"),
            QLabel("Volume (stk voxel)"),
            QLabel("Proportion (%)"),
            QLabel("Color"),
        ]
        for i in range(len(column_names)):
            sec_layout.addWidget(column_names[i], 0, i, 1, 1)

        for i in range(len(region_label)):
            clb = QLabel()
            da_color = QColor(
                region_color[i][0], region_color[i][1], region_color[i][2], 255
            ).name()
            clb.setStyleSheet(
                "QLabel {background-color: " + da_color + "; width: 20px; height: 20px}"
            )

            row_val = [
                QLabel(str(region_label[i])),
                QLabel(region_name[i]),
                QLabel(region_acronym[i]),
                QLabel(str(virus_volume[i])),
                QLabel(str(proportion[i])),
                clb,
            ]

            for j in range(len(row_val)):
                sec_layout.addWidget(row_val[j], i + 1, j, 1, 1)

        # ok button, used to close window
        ok_btn = QDialogButtonBox(QDialogButtonBox.Ok)
        ok_btn.accepted.connect(self.accept)

        # add widget to layout
        layout.addWidget(self.label)
        layout.addWidget(sec_group)
        layout.addWidget(ok_btn)
        self.setLayout(layout)

    def accept(self) -> None:
        self.close()

    def set_probe_color(self, color):
        self.label.setStyleSheet("QLabel {background-color: " + color + ";}")


class ProbeInfoWindow(QDialog):
    def __init__(self, name, data):
        super().__init__()

        self.setWindowTitle("Probe Information Window")

        self.label = QLabel(name)
        # self.label = QLabel("Probe % d " % group_id)
        color = QColor(
            data["vis_color"][0],
            data["vis_color"][1],
            data["vis_color"][2],
            data["vis_color"][3],
        )
        label_style = "QLabel {background-color: " + color.name() + "; font-size: 20px}"
        self.label.setStyleSheet(label_style)

        ap_angle_label = QLabel("AP Angle : ")
        ap_angle_label.setAlignment(Qt.AlignCenter)
        ml_angle_label = QLabel("ML Angle : ")
        ml_angle_label.setAlignment(Qt.AlignCenter)
        probe_length_label = QLabel("Probe Length : ")
        probe_length_label.setAlignment(Qt.AlignCenter)
        dv_label = QLabel("DV : ")
        dv_label.setAlignment(Qt.AlignCenter)
        insertion_coords_label = QLabel("Insertion coordinates : ")
        insertion_coords_label.setAlignment(Qt.AlignCenter)
        insertion_voxels_label = QLabel("Insertion voxels : ")
        insertion_voxels_label.setAlignment(Qt.AlignCenter)

        terminus_coords_label = QLabel("Terminus coordinates : ")
        terminus_coords_label.setAlignment(Qt.AlignCenter)
        terminus_voxels_label = QLabel("Terminus voxels : ")
        terminus_voxels_label.setAlignment(Qt.AlignCenter)

        ap_angle_value = QLabel("{} \u00B0".format(np.round(data["ap_angle"], 2)))
        ml_angle_value = QLabel("{} \u00B0".format(np.round(data["ml_angle"], 2)))
        probe_length = QLabel("{} \u03BCm".format(np.round(data["probe_length"], 2)))
        dv = QLabel("{} \u03BCm".format(np.round(data["dv"], 2)))
        ic_val = np.round(data["insertion_coords"], 2)
        insertion_coords = QLabel(
            "ML: {}\u03BCm,  AP: {}\u03BCm".format(ic_val[0], ic_val[1])
        )
        iv_val = data["insertion_vox"].astype(int)
        insertion_vox = QLabel("({} {} {})".format(iv_val[0], iv_val[1], iv_val[2]))

        tc_val = np.round(data["terminus_coords"], 2)
        terminus_coords = QLabel(
            "ML: {}\u03BCm,  AP: {}\u03BCm".format(tc_val[0], tc_val[1])
        )
        tv_val = data["terminus_vox"]
        terminus_vox = QLabel("({} {} {})".format(tv_val[0], tv_val[1], tv_val[2]))

        coords_info_group = QGroupBox()
        coords_info_layout = QGridLayout(coords_info_group)
        coords_info_layout.addWidget(ap_angle_label, 0, 0, 1, 1)
        coords_info_layout.addWidget(ap_angle_value, 0, 1, 1, 1)
        coords_info_layout.addWidget(ml_angle_label, 0, 2, 1, 1)
        coords_info_layout.addWidget(ml_angle_value, 0, 3, 1, 1)

        coords_info_layout.addWidget(probe_length_label, 1, 0, 1, 1)
        coords_info_layout.addWidget(probe_length, 1, 1, 1, 1)
        coords_info_layout.addWidget(dv_label, 1, 2, 1, 1)
        coords_info_layout.addWidget(dv, 1, 3, 1, 1)

        coords_info_layout.addWidget(insertion_coords_label, 2, 0, 1, 1)
        coords_info_layout.addWidget(insertion_coords, 2, 1, 1, 1)
        coords_info_layout.addWidget(terminus_coords_label, 2, 2, 1, 1)
        coords_info_layout.addWidget(terminus_coords, 2, 3, 1, 1)

        coords_info_layout.addWidget(insertion_voxels_label, 3, 0, 1, 1)
        coords_info_layout.addWidget(insertion_vox, 3, 1, 1, 1)
        coords_info_layout.addWidget(terminus_voxels_label, 3, 2, 1, 1)
        coords_info_layout.addWidget(terminus_vox, 3, 3, 1, 1)

        # group info container
        region_sites_num = np.ravel(data["region_sites"]).astype(str)[::-1]
        region_label = np.ravel(data["region_label"]).astype(int).astype(str)[::-1]
        region_color = np.asarray(data["label_color"])[::-1, :]
        region_length = np.round(data["region_length"], 3).astype(str)[::-1]
        region_name = np.ravel(data["label_name"])[::-1]
        region_acronym = np.ravel(data["label_acronym"])[::-1]

        sec_group = QGroupBox()
        sec_layout = QGridLayout(sec_group)
        column_names = [
            QLabel("ID"),
            QLabel("Brain Region"),
            QLabel("Acronym"),
            QLabel("Sites (stk)"),
            QLabel("Length (um)"),
            QLabel("Color"),
        ]
        for i in range(len(column_names)):
            sec_layout.addWidget(column_names[i], 0, i, 1, 1)

        for i in range(len(region_label)):
            clb = QLabel()
            da_color = QColor(
                region_color[i][0], region_color[i][1], region_color[i][2], 255
            ).name()
            clb.setStyleSheet(
                "QLabel {background-color: " + da_color + "; width: 20px; height: 20px}"
            )

            row_val = [
                QLabel(region_label[i]),
                QLabel(region_name[i]),
                QLabel(region_acronym[i]),
                QLabel(region_sites_num[i]),
                QLabel(region_length[i]),
                clb,
            ]

            for j in range(len(row_val)):
                sec_layout.addWidget(row_val[j], i + 1, j, 1, 1)

        # plot container
        plot_frame = QFrame()
        plot_frame.setMaximumWidth(300)
        plot_frame.setMinimumWidth(300)
        view_layout = QHBoxLayout(plot_frame)
        view_layout.setSpacing(0)
        view_layout.setContentsMargins(0, 0, 0, 0)

        w = pg.GraphicsLayoutWidget()
        view = w.addViewBox()
        view.setAspectLocked()
        # view.invertY(True)

        # load plot data
        vis_data = data["vis_data"]
        n_column = len(vis_data)

        probe_type_name = data["probe_type_name"]
        # draw tips
        if probe_type_name != "Tetrode":
            da_tip_loc = np.array(
                [[0, 0], [0.5 * n_column, -2 * n_column], [n_column, 0]]
            )
            tips = pg.PlotDataItem(
                da_tip_loc,
                connect="all",
                fillLevel=0,
                fillBrush=pg.mkBrush(color=(128, 128, 128)),
            )
            view.addItem(tips)

        # draw area
        region_label = data["region_label"]
        region_sites = data["region_sites"]
        region_text_loc = data["text_loc"]

        group_color = data["label_color"]
        sites_label = data["sites_label"]
        for i in range(n_column):
            group_id = vis_data[i]["group_id"].astype(int)
            start_loc = vis_data[i]["start_loc"]
            end_loc = vis_data[i]["end_loc"]
            sites_loc_column = vis_data[i]["sites"]

            for j in range(len(group_id)):
                area_data = np.array([[i, end_loc[j]], [i + 1, end_loc[j]]])
                da_item = pg.PlotDataItem(
                    area_data,
                    fillLevel=start_loc[j],
                    fillBrush=group_color[group_id[j]],
                    pen=None,
                )
                view.addItem(da_item)

            sites_color = []
            for j in range(len(sites_label[i])):
                color_ind = np.where(region_label == sites_label[i][j])[0][0]
                sites_color.append(group_color[color_ind])

            pnt_data = np.stack(
                [np.repeat(i + 0.5, len(sites_loc_column)), sites_loc_column], axis=1
            )
            da_item = pg.ScatterPlotItem(
                pos=pnt_data,
                pen=(128, 128, 128, 128),
                brush=sites_color,
                symbol="s",
                size=3,
                hoverSize=8,
            )
            view.addItem(da_item)

        # draw text
        for i in range(len(region_label)):
            region_text = pg.TextItem(str(region_sites[i]))
            region_text.setPos(n_column + 0.5, region_text_loc[i])
            view.addItem(region_text)

        view_layout.addWidget(w)

        channel_info_frame = QFrame()
        channel_info_layout = QHBoxLayout(channel_info_frame)
        channel_info_layout.setContentsMargins(0, 0, 0, 0)
        channel_info_layout.setSpacing(10)
        channel_info_layout.addWidget(plot_frame)
        channel_info_layout.addWidget(sec_group)

        # ok button, used to close window
        ok_btn = QDialogButtonBox(QDialogButtonBox.Ok)
        ok_btn.accepted.connect(self.accept)

        # add widget to layout
        layout = QVBoxLayout()
        layout.addWidget(self.label)
        layout.addWidget(coords_info_group)
        layout.addSpacing(10)
        layout.addWidget(channel_info_frame)
        layout.addWidget(ok_btn)
        self.setLayout(layout)

    def accept(self) -> None:
        self.close()

    def set_probe_color(self, color):
        self.label.setStyleSheet("QLabel {background-color: " + color + ";}")


class SinglePiece(QWidget):
    sig_clicked = pyqtSignal(object)
    sig_name_changed = pyqtSignal(object)

    def __init__(
        self,
        parent=None,
        index=0,
        obj_name="",
        obj_type="probe piece",
        object_icon=None,
    ):
        QWidget.__init__(self, parent=parent)

        self.inactive_style = "QFrame{background-color:rgb(83, 83, 83); border: 1px solid rgb(128, 128, 128);}"
        self.active_style = "QFrame{background-color:rgb(107, 107, 107); border: 1px solid rgb(128, 128, 128);}"

        self.id = index
        self.object_name = obj_name
        self.object_type = obj_type
        self.active = True

        self.tbnail = QPushButton()
        self.tbnail.setFixedSize(QSize(40, 40))
        self.tbnail.setStyleSheet(eye_button_style)
        self.tbnail.setIcon(object_icon)
        self.tbnail.setIconSize(QSize(20, 20))
        self.tbnail.clicked.connect(self.on_click)

        self.text_btn = QDoubleButton()
        self.text_btn.setFixedSize(QSize(240, 40))
        self.text_btn.left_clicked.connect(self.on_click)
        self.text_btn.double_clicked.connect(self.on_doubleclick)

        self.l_line_edit = QLineEdit()
        self.l_line_edit.setStyleSheet(line_edit_style)
        self.l_line_edit.setFixedWidth(240)
        self.l_line_edit.editingFinished.connect(self.enter_pressed)
        self.l_line_edit.setVisible(False)

        self.inner_frame = QFrame()
        self.inner_frame.setStyleSheet(self.active_style)
        self.inner_layout = QHBoxLayout(self.inner_frame)
        self.inner_layout.setContentsMargins(0, 0, 0, 0)
        self.inner_layout.setSpacing(0)
        self.inner_layout.setAlignment(Qt.AlignVCenter)
        self.inner_layout.addWidget(self.tbnail)
        self.inner_layout.addSpacing(5)
        self.inner_layout.addWidget(self.text_btn)
        self.inner_layout.addWidget(self.l_line_edit)
        self.inner_layout.addStretch()

        outer_layout = QHBoxLayout()
        outer_layout.setContentsMargins(0, 0, 0, 0)
        outer_layout.setSpacing(0)
        outer_layout.setAlignment(Qt.AlignVCenter)
        outer_layout.addWidget(self.inner_frame)

        self.setLayout(outer_layout)
        self.setFixedHeight(40)
        # self.show()

    @pyqtSlot()
    def on_click(self):
        self.set_checked(True)
        self.sig_clicked.emit(self.id)

    @pyqtSlot()
    def on_doubleclick(self):
        self.l_line_edit.setText(self.text_btn.text())
        self.l_line_edit.setVisible(True)
        self.text_btn.setVisible(False)
        self.l_line_edit.setFocus(True)
        self.set_checked(True)
        self.sig_clicked.emit(self.id)

    @pyqtSlot()
    def enter_pressed(self):
        da_text = self.l_line_edit.text()
        if "-" not in da_text or "piece" not in da_text:
            return
        self.l_line_edit.setVisible(False)
        self.text_btn.setText(self.l_line_edit.text())
        self.object_name = self.l_line_edit.text()
        self.text_btn.setVisible(True)
        self.sig_name_changed.emit((self.id, self.object_name))

    def is_checked(self):
        return self.active

    def set_checked(self, check):
        self.active = check
        if not self.active:
            self.inner_frame.setStyleSheet(self.inactive_style)
        else:
            self.inner_frame.setStyleSheet(self.active_style)


class RegisteredObject(QWidget):
    sig_object_color_changed = pyqtSignal(object)
    eye_clicked = pyqtSignal(object)
    sig_clicked = pyqtSignal(object)
    sig_link = pyqtSignal(object)
    sig_double_clicked = pyqtSignal(object)

    def __init__(
        self,
        parent=None,
        obj_id=0,
        obj_name="",
        obj_type="merged probe",
        object_icon=None,
    ):
        QWidget.__init__(self, parent=parent)

        self.inactive_style = "QFrame{background-color:rgb(83, 83, 83); border: 1px solid rgb(128, 128, 128);}"
        self.active_style = "QFrame{background-color:rgb(107, 107, 107); border: 1px solid rgb(128, 128, 128);}"

        self.color = QColor(randint(0, 255), randint(0, 255), randint(0, 255), 255)
        self.icon_back = "border:1px solid black; background-color: {}".format(
            self.color.name()
        )

        self.id = obj_id
        self.object_type = obj_type
        self.object_name = obj_name
        self.active = True
        self.vis = True

        self.eye_button = QPushButton()
        self.eye_button.setFixedSize(QSize(40, 40))
        self.eye_button.setStyleSheet(eye_button_style)
        self.eye_button.setCheckable(True)
        eye_icon = QIcon()
        eye_icon.addPixmap(QPixmap("icons/layers/eye_on.png"), QIcon.Normal, QIcon.Off)
        eye_icon.addPixmap(QPixmap("icons/layers/eye_off.png"), QIcon.Normal, QIcon.On)
        self.eye_button.setIcon(eye_icon)
        self.eye_button.setIconSize(QSize(20, 20))
        self.eye_button.clicked.connect(self.eye_on_click)

        self.tbnail = QPushButton()
        self.tbnail.setStyleSheet(self.icon_back)
        self.tbnail.setIcon(object_icon)
        self.tbnail.setIconSize(QSize(20, 20))
        self.tbnail.clicked.connect(self.change_object_color)

        self.text_btn = ObjectTextButton()
        self.text_btn.setText(self.object_name)
        self.text_btn.clicked.connect(self.text_btn_on_click)
        # self.text_btn.setSizePolicy(QSizePolicy.Expanding, QSizePolicy.Ignored)

        self.link_button = QPushButton()
        self.link_button.setFixedSize(QSize(25, 40))
        self.link_button.setCheckable(True)
        link_icon = QIcon()
        link_icon.addPixmap(
            QPixmap("icons/sidebar/link_off.svg"), QIcon.Normal, QIcon.Off
        )
        link_icon.addPixmap(QPixmap("icons/sidebar/link.svg"), QIcon.Normal, QIcon.On)
        self.link_button.setIcon(link_icon)
        self.link_button.setIconSize(QSize(20, 20))
        self.link_button.clicked.connect(self.on_linked)

        self.inner_frame = QFrame()
        self.inner_frame.setStyleSheet(self.active_style)
        self.inner_layout = QHBoxLayout(self.inner_frame)
        self.inner_layout.setContentsMargins(0, 0, 0, 0)
        self.inner_layout.setSpacing(0)
        self.inner_layout.setAlignment(Qt.AlignVCenter)
        self.inner_layout.addWidget(self.eye_button)
        self.inner_layout.addSpacing(5)
        self.inner_layout.addWidget(self.tbnail)
        self.inner_layout.addSpacing(5)
        self.inner_layout.addWidget(self.text_btn)
        self.inner_layout.addWidget(self.link_button)
        self.inner_layout.addStretch()

        outer_layout = QHBoxLayout()
        outer_layout.setContentsMargins(0, 0, 0, 0)
        outer_layout.setSpacing(0)
        outer_layout.setAlignment(Qt.AlignVCenter)
        outer_layout.addWidget(self.inner_frame)

        self.setLayout(outer_layout)
        self.setFixedHeight(40)

    def set_icon_style(self, color):
        self.color = color
        self.icon_back = "border:1px solid black; background-color: {}".format(
            color.name()
        )
        self.tbnail.setStyleSheet(self.icon_back)

    def change_object_color(self):
        self.sig_object_color_changed.emit(self.id)

    def eye_on_click(self):
        if self.eye_button.isChecked():
            self.vis = False
        else:
            self.vis = True
        self.set_checked(True)
        self.tbnail.setEnabled(self.vis)
        self.text_btn.setEnabled(self.vis)
        self.eye_clicked.emit((self.id, self.vis))

    def set_checked(self, check):
        self.active = check
        if not self.active:
            self.inner_frame.setStyleSheet(self.inactive_style)
        else:
            self.inner_frame.setStyleSheet(self.active_style)

    def text_btn_on_click(self):
        self.set_checked(True)
        self.sig_clicked.emit(self.id)

    def is_checked(self):
        return self.active

    def on_linked(self):
        self.sig_link.emit(self.id)


class BottomButton(QPushButton):
    def __init__(self, icon_path):
        QPushButton.__init__(self)
        btm_style = read_qss_file("qss/obj_ctrl_bottom_button.qss")
        self.setFixedSize(24, 24)
        self.setStyleSheet(btm_style)
        self.setIcon(QIcon(icon_path))
        self.setIconSize(QSize(20, 20))


class ObjectControl(QObject):
    """
    only pieces' name can be changed, the merged can not, so far
    """

    class SignalProxy(QObject):
        sigOpacityChanged = pyqtSignal(object)
        sigVisChanged = pyqtSignal(object)
        sigDeleteObject = pyqtSignal(object)
        sigAddObject = pyqtSignal(object)
        sigColorChanged = pyqtSignal(object)
        sigSizeChanged = pyqtSignal(object)
        sigBlendModeChanged = pyqtSignal(object)

    def __init__(self, parent=None):
        self._sigprox = ObjectControl.SignalProxy()
        self.sig_opacity_changed = self._sigprox.sigOpacityChanged
        self.sig_visible_changed = self._sigprox.sigVisChanged
        self.sig_delete_object = (
            self._sigprox.sigDeleteObject
        )  # for delete obj 3d gl widgets
        self.sig_add_object = self._sigprox.sigAddObject  # for add obj 3d gl widgets
        self.sig_color_changed = self._sigprox.sigColorChanged
        self.sig_size_changed = self._sigprox.sigSizeChanged
        self.sig_blend_mode_changed = self._sigprox.sigBlendModeChanged

        QObject.__init__(self)

        self.default_size_val = 2
        self.default_opacity_val = 100

        self.valid_obj_type = [
            "probe piece",
            "merged probe",
            "cells piece",
            "merged cells",
            "virus piece",
            "merged virus",
            "contour piece",
            "merged contour",
            "drawing piece",
            "merged drawing",
        ]

        self.current_obj_index = None

        self.obj_count = 0
        self.linked_indexes = []

        self.obj_list = []  # widgets
        self.obj_id = []  # identity
        self.obj_name = (
            []
        )  # names, initially the same as type, can be changed freely ???
        self.obj_type = []  # type
        self.obj_data = []  # data
        self.obj_size = []  # size
        self.obj_opacity = []
        self.obj_comp_mode = []
        self.obj_visibility = []
        self.obj_merged = []

        self.probe_icon = QIcon("icons/sidebar/probe.svg")
        self.virus_icon = QIcon("icons/sidebar/virus.svg")
        self.drawing_icon = QIcon("icons/toolbar/pencil.svg")
        self.cell_icon = QIcon("icons/toolbar/location.svg")
        self.contour_icon = QIcon("icons/sidebar/contour.svg")
        self.compare_icon = QIcon("icons/sidebar/compare.svg")

        combo_label = QLabel("Composition:")
        combo_label.setFixedWidth(80)
        self.obj_blend_combo = QComboBox()
        self.obj_blend_combo.setEditable(False)
        combo_value = ["opaque", "translucent", "additive"]
        self.obj_blend_combo.addItems(combo_value)
        self.obj_blend_combo.setCurrentText("opaque")
        self.obj_blend_combo.currentTextChanged.connect(self.blend_mode_changed)
        combo_wrap = QFrame()
        combo_wrap_layout = QHBoxLayout(combo_wrap)
        combo_wrap_layout.setContentsMargins(0, 0, 0, 0)
        combo_wrap_layout.setSpacing(5)
        combo_wrap_layout.addWidget(combo_label)
        combo_wrap_layout.addWidget(self.obj_blend_combo)

        obj_opacity_label = QLabel("Opacity:")
        obj_opacity_label.setFixedWidth(80)
        self.obj_opacity_slider = QSlider(Qt.Horizontal)
        self.obj_opacity_slider.setMaximum(100)
        self.obj_opacity_slider.setMinimum(0)
        self.obj_opacity_slider.setValue(100)
        self.obj_opacity_slider.valueChanged.connect(self.change_opacity_label_value)
        self.obj_opacity_slider.sliderMoved.connect(self.send_opacity_changed_signal)
        self.obj_opacity_val_label = QLabel("100%")
        self.obj_opacity_val_label.setFixedWidth(40)
        opacity_wrap = QFrame()
        opacity_wrap_layout = QHBoxLayout(opacity_wrap)
        opacity_wrap_layout.setContentsMargins(0, 0, 0, 0)
        opacity_wrap_layout.setSpacing(5)
        opacity_wrap_layout.addWidget(obj_opacity_label)
        opacity_wrap_layout.addWidget(self.obj_opacity_slider)
        opacity_wrap_layout.addWidget(self.obj_opacity_val_label)

        obj_size_label = QLabel("Size/Width: ")
        obj_size_label.setFixedWidth(80)
        self.obj_size_slider = QSlider(Qt.Horizontal)
        self.obj_size_slider.setValue(2)
        self.obj_size_slider.setMinimum(1)
        self.obj_size_slider.setMaximum(10)
        self.obj_size_slider.valueChanged.connect(self.change_size_label_value)
        self.obj_size_slider.sliderMoved.connect(self.send_size_changed_signal)
        self.obj_size_val_label = QLabel("2")
        self.obj_size_val_label.setAlignment(Qt.AlignCenter)
        self.obj_size_val_label.setFixedWidth(40)
        size_wrap = QFrame()
        size_wrap_layout = QHBoxLayout(size_wrap)
        size_wrap_layout.setContentsMargins(0, 0, 0, 0)
        size_wrap_layout.setSpacing(5)
        size_wrap_layout.addWidget(obj_size_label)
        size_wrap_layout.addWidget(self.obj_size_slider)
        size_wrap_layout.addWidget(self.obj_size_val_label)

        top_frame = QFrame()
        top_layout = QVBoxLayout(top_frame)
        top_layout.setContentsMargins(0, 0, 0, 0)
        top_layout.setSpacing(0)
        top_layout.addWidget(combo_wrap)
        top_layout.addSpacing(10)
        top_layout.addWidget(opacity_wrap)
        top_layout.addSpacing(10)
        top_layout.addWidget(size_wrap)
        top_layout.addSpacing(10)

        self.layer_frame = QFrame()
        self.layer_frame.setStyleSheet("background: transparent; border: 0px;")
        self.layer_frame.setSizePolicy(QSizePolicy.Expanding, QSizePolicy.Expanding)
        self.layer_layout = QBoxLayout(QBoxLayout.BottomToTop, self.layer_frame)
        self.layer_layout.setAlignment(Qt.AlignBottom)
        self.layer_layout.setContentsMargins(0, 0, 0, 0)
        self.layer_layout.setSpacing(5)

        self.layer_scroll = QScrollArea()
        self.layer_scroll.setStyleSheet("background: transparent; border: 0px;")
        self.layer_scroll.setSizePolicy(QSizePolicy.Expanding, QSizePolicy.Expanding)
        self.layer_scroll.setWidget(self.layer_frame)
        self.layer_scroll.setVerticalScrollBarPolicy(Qt.ScrollBarAlwaysOn)
        self.layer_scroll.setHorizontalScrollBarPolicy(Qt.ScrollBarAlwaysOff)
        self.layer_scroll.setWidgetResizable(True)

        mid_frame = QFrame()
        mid_frame.setStyleSheet(
            "background: transparent; border: 1px solid rgb(128, 128, 128);"
        )
        mid_frame.setSizePolicy(QSizePolicy.Expanding, QSizePolicy.Expanding)
        mid_layout = QGridLayout(mid_frame)
        mid_layout.setContentsMargins(0, 0, 0, 0)
        mid_layout.setSpacing(0)
        mid_layout.setAlignment(Qt.AlignBottom)
        mid_layout.addWidget(self.layer_scroll, 0, 0, 1, 1)

        self.outer_frame = QFrame()
        self.outer_frame.setSizePolicy(QSizePolicy.Expanding, QSizePolicy.Expanding)
        outer_layout = QVBoxLayout(self.outer_frame)
        outer_layout.setSpacing(0)
        outer_layout.addWidget(top_frame)
        outer_layout.addWidget(mid_frame)

        # bottom buttons
        self.info_btn = BottomButton("icons/sidebar/info.svg")
        self.info_btn.setToolTip("open information window")
        self.info_btn.clicked.connect(self.info_btn_clicked)

        self.vis2d_btn = BottomButton("icons/toolbar/vis2d.svg")
        self.vis2d_btn.setToolTip("find the corresponding 2D plane")
        self.unmerge_btn = BottomButton("icons/toolbar/unmerge.svg")
        self.unmerge_btn.setToolTip("un-merge a merged object")
        self.unmerge_btn.clicked.connect(self.unmerge_objects)
        self.compare_btn = BottomButton("icons/sidebar/compare.svg")
        self.compare_btn.setToolTip("compare multiple linked objects")

        self.merge_probe_btn = BottomButton("icons/sidebar/probe.svg")
        self.merge_probe_btn.setToolTip("merge probes pieces")

        self.merge_drawing_btn = BottomButton("icons/toolbar/pencil.svg")
        self.merge_drawing_btn.setToolTip("merge drawing pieces")

        self.merge_cell_btn = BottomButton("icons/toolbar/location.svg")
        self.merge_cell_btn.setToolTip("merge cells pieces")

        self.merge_virus_btn = BottomButton("icons/sidebar/virus.svg")
        self.merge_virus_btn.setToolTip("merge virus pieces")

        self.merge_contour_btn = BottomButton("icons/sidebar/contour.svg")
        self.merge_contour_btn.setToolTip("merge contour pieces")

        self.add_object_btn = BottomButton("icons/sidebar/add.png")
        self.add_object_btn.setToolTip("add a piece")

        self.delete_object_btn = BottomButton("icons/sidebar/trash.png")
        self.delete_object_btn.setToolTip("delete an object")
        self.delete_object_btn.clicked.connect(self.delete_object_btn_clicked)

    def blend_mode_changed(self):
        blend_mode = self.obj_blend_combo.currentText()
        if self.current_obj_index is None:
            return
        if "merged" not in self.obj_type[self.current_obj_index]:
            return
        if blend_mode != self.obj_comp_mode[self.current_obj_index]:
            self.obj_comp_mode[self.current_obj_index] = blend_mode
            self.sig_blend_mode_changed.emit(blend_mode)

    def change_opacity_label_value(self):
        da_val = self.obj_opacity_slider.value()
        self.obj_opacity_val_label.setText("{} %".format(da_val))

    def send_opacity_changed_signal(self):
        da_val = self.obj_opacity_slider.value()
        if self.current_obj_index is None:
            return
        if "merged" not in self.obj_type[self.current_obj_index]:
            self.obj_opacity_slider.setValue(da_val)
            return
        self.obj_opacity[self.current_obj_index] = da_val
        self.sig_opacity_changed.emit(da_val)

    def change_size_label_value(self):
        da_val = self.obj_size_slider.value()
        self.obj_size_val_label.setText(str(da_val))

    def send_size_changed_signal(self):
        da_val = self.obj_size_slider.value()
        if self.current_obj_index is None:
            return
        if "merged" not in self.obj_type[self.current_obj_index]:
            self.obj_size_slider.setValue(da_val)
            return
        self.obj_size[self.current_obj_index] = da_val
        self.sig_size_changed.emit((self.obj_type[self.current_obj_index], da_val))

    def obj_piece_name_changed(self, ev):
        clicked_id = ev[0]
        name = ev[1]
        clicked_index = np.where(np.ravel(self.obj_id) == clicked_id)[0][0]
        self.current_obj_index = clicked_index
        self.obj_name[clicked_index] = name

    def delete_object_btn_clicked(self):
        if self.current_obj_index is None:
            return
        self.delete_objects(self.current_obj_index)

    def info_btn_clicked(self):
        if "merged" in self.obj_type[self.current_obj_index]:
            self.obj_info_on_click()

    def obj_clicked(self, clicked_id):
        self.set_active_layer_to_current(clicked_id)

    def obj_info_on_click(self):
        da_data = self.obj_data[self.current_obj_index]
        da_name = self.obj_name[self.current_obj_index]
        if "probe" in self.obj_type[self.current_obj_index]:
            self.info_window = ProbeInfoWindow(da_name, da_data)
        elif "virus" in self.obj_type[self.current_obj_index]:
            self.info_window = VirusInfoWindow(da_name, da_data)
        elif "cell" in self.obj_type[self.current_obj_index]:
            self.info_window = CellsInfoWindow(da_name, da_data)
        else:
            return
        self.info_window.exec()

    def set_active_layer_to_current(self, clicked_id):
        previous_obj_id = self.obj_id[self.current_obj_index]
        if previous_obj_id != clicked_id:
            active_index = np.where(np.ravel(self.obj_id) == clicked_id)[0][0]
            self.current_obj_index = active_index
            # self.obj_list[active_index].set_checked(True)
            inactive_id = np.where(np.ravel(self.obj_id) == previous_obj_id)[0][0]
            self.obj_list[inactive_id].set_checked(False)
            self.set_slider_combo_to_current()
        else:
            return

    def obj_link_changed(self, clicked_id):
        da_index = np.where(np.ravel(self.obj_id) == clicked_id)[0][0]
        if self.obj_list[da_index].link_button.isChecked():
            self.linked_indexes.append(da_index)
        else:
            if da_index in self.linked_indexes:
                del_ind = np.where(np.ravel(self.linked_indexes) == da_index)[0][0]
                self.linked_indexes.pop(del_ind)

    def obj_color_changed(self, clicked_id):
        self.set_active_layer_to_current(clicked_id)
        color = QColorDialog.getColor()
        if color.isValid():
            self.obj_list[self.current_obj_index].set_icon_style(color)
            da_color = (color.red(), color.green(), color.blue(), 255)
            self.sig_color_changed.emit((self.current_obj_index, da_color))

    def obj_eye_clicked(self, ev):
        clicked_id = ev[0]
        vis = ev[1]
        self.set_active_layer_to_current(clicked_id)
        self.sig_visible_changed.emit((self.current_obj_index, vis))

    # delete a object
    def delete_objects(self, delete_index):
        del_ind = np.ravel(delete_index)
        if len(del_ind) > 1:
            del_ind = np.sort(delete_index)[::-1]

        current_obj_id = self.obj_id[self.current_obj_index]
        for da_ind in del_ind:
            self.layer_layout.removeWidget(self.obj_list[da_ind])
            self.obj_list[da_ind].deleteLater()
            del self.obj_list[da_ind]
            del self.obj_id[da_ind]
            del self.obj_name[da_ind]
            del self.obj_type[da_ind]
            del self.obj_data[da_ind]
            del self.obj_comp_mode[da_ind]
            del self.obj_opacity[da_ind]
            del self.obj_size[da_ind]
            self.sig_delete_object.emit(da_ind)
        if self.current_obj_index in del_ind:
            if self.obj_list:
                self.obj_list[-1].set_checked(True)
                self.current_obj_index = len(self.obj_list) - 1
            else:
                self.current_obj_index = None
        else:
            active_index = np.where(np.ravel(self.obj_id) == current_obj_id)[0][0]
            self.current_obj_index = active_index

    #
    def get_object_icon(self, object_type):
        if "probe" in object_type:
            object_icon = self.probe_icon
        elif "virus" in object_type:
            object_icon = self.virus_icon
        elif "cell" in object_type:
            object_icon = self.cell_icon
        elif "contour" in object_type:
            object_icon = self.contour_icon
        elif "drawing" in object_type:
            object_icon = self.drawing_icon
        else:
            object_icon = None
        return object_icon

    def get_group_count(self, object_type):
        group_count = len(
            [da_type for da_type in self.obj_type if da_type == object_type]
        )
        return group_count

    def add_object(self, object_name, object_type, object_data, object_mode):
        object_icon = self.get_object_icon(object_type)
        # group_count = self.get_group_count(object_type)
        if object_icon is None:
            return
        self.obj_id.append(self.obj_count)
        self.obj_data.append(object_data)
        self.obj_name.append(object_name)
        self.obj_type.append(object_type)
        if "merged" in object_type:
            new_layer = RegisteredObject(
                obj_id=self.obj_count,
                obj_name=object_name,
                obj_type=object_type,
                object_icon=object_icon,
            )
            new_layer.eye_clicked.connect(self.obj_eye_clicked)
            new_layer.sig_object_color_changed.connect(self.obj_color_changed)
            new_layer.sig_link.connect(self.obj_link_changed)
            self.obj_opacity.append(self.default_opacity_val)
            self.obj_size.append(self.default_size_val)
            self.obj_comp_mode.append(object_mode)
            da_color = (
                new_layer.color.red(),
                new_layer.color.green(),
                new_layer.color.blue(),
                255,
            )
            self.obj_data[-1]["vis_color"] = da_color
            if self.obj_size_slider.value() != self.default_size_val:
                self.obj_size_slider.setValue(self.default_size_val)
            if self.obj_opacity_slider.value() != self.default_opacity_val:
                self.obj_opacity_slider.setValue(self.default_opacity_val)
            if self.obj_blend_combo.currentText() != object_mode:
                self.obj_blend_combo.blockSignals(True)
                self.obj_blend_combo.setCurrentText(object_mode)
                self.obj_blend_combo.blockSignals(False)
        else:
            new_layer = SinglePiece(
                index=self.obj_count,
                obj_name=object_name,
                obj_type=object_type,
                object_icon=object_icon,
            )
            new_layer.sig_name_changed.connect(self.obj_piece_name_changed)
            self.obj_opacity.append([])
            self.obj_size.append([])
            self.obj_comp_mode.append([])

        new_layer.text_btn.setText(self.obj_name[-1])
        new_layer.set_checked(True)
        new_layer.sig_clicked.connect(self.obj_clicked)
        self.obj_list.append(new_layer)

        active_index = np.where(np.ravel(self.obj_id) == self.obj_count)[0][0]
        if self.current_obj_index is None:
            self.current_obj_index = active_index
        else:
            self.obj_list[self.current_obj_index].set_checked(False)
            self.current_obj_index = active_index

        self.layer_layout.addWidget(self.obj_list[-1])
        self.sig_add_object.emit((object_data, object_type))
        self.obj_count += 1

    # merge object pieces
    def merge_pieces(self, obj_type="probe piece"):
        if obj_type not in [
            "probe piece",
            "virus piece",
            "contour piece",
            "drawing piece",
            "cells piece",
        ]:
            return
        n_obj = len(self.obj_id)
        cind = [ind for ind in range(n_obj) if self.obj_type[ind] == obj_type]
        # print("cind", cind)
        valid_pieces_names = np.ravel(self.obj_name)[np.array(cind)]
        n_pieces = len(valid_pieces_names)
        m_obj_names = []
        for i in range(n_pieces):
            da_name = valid_pieces_names[i]
            # da_name = da_name.replace(" ", "")
            da_name_split = da_name.split("-")
            m_obj_name = da_name_split[0]
            if m_obj_name[-1] == " ":
                m_obj_name = m_obj_name[:-1]
            m_obj_names.append(m_obj_name)
        merging_object_names = np.unique(m_obj_names)
        n_object = len(merging_object_names)
        data = [[] for _ in range(n_object)]
        pieces_names = [[] for _ in range(n_object)]
        for i in range(n_object):
            sub_inds = [
                cind[ind]
                for ind in range(n_pieces)
                if m_obj_names[ind] == merging_object_names[i]
            ]
            for j in range(len(sub_inds)):
                data[i].append(self.obj_data[sub_inds[j]])
                pieces_names[i].append(self.obj_name[sub_inds[j]])
            # temp = self.obj_data[da_piece_ind_in_obj_order[0]].T
            # if len(da_piece_ind_in_obj_order) > 1:
            #     for j in range(1, len(da_piece_ind_in_obj_order)):
            #         print(self.obj_data[da_piece_ind_in_obj_order[j]].T)
            #         temp = np.hstack([temp, self.obj_data[da_piece_ind_in_obj_order[j]].T])
            # data[i] = temp.T
        self.delete_objects(cind)
        return data, merging_object_names, pieces_names

    # unmerge object pieces
    def unmerge_objects(self):
        current_type = self.obj_type[self.current_obj_index]
        if "merged" not in current_type:
            return
        current_data = self.obj_data[self.current_obj_index]
        m_obj_type = current_type.split(" ")[1]
        data_list = current_data["data"]
        pieces_names = current_data["pieces_names"]
        self.delete_objects([self.current_obj_index])
        for i in range(len(data_list)):
            self.add_object(
                pieces_names[i], "{} piece".format(m_obj_type), data_list[i], None
            )

    # get obj data
    def get_obj_data(self):
        data = {
            "obj_name": self.obj_name,
            "obj_type": self.obj_type,
            "obj_data": self.obj_data,
            "obj_size": self.obj_size,
            "obj_opacity": self.obj_opacity,
            "obj_comp_mode": self.obj_comp_mode,
            "current_obj_index": self.current_obj_index,
        }
        return data

    def set_obj_data(self, data):
        for i in range(len(data["obj_type"])):
            self.add_object(
                data["obj_name"][i],
                data["obj_type"][i],
                data["obj_data"][i],
                data["obj_comp_mode"][i],
            )

        self.obj_size = data["obj_size"]
        self.obj_opacity = data["obj_opacity"]
        self.obj_comp_mode = data["obj_comp_mode"]
        self.current_obj_index = data["current_obj_index"]

        self.obj_list[-1].set_checked(False)
        self.obj_list[self.current_obj_index].set_checked(True)

        # print(self.obj_type)

    def set_slider_combo_to_current(self):
        if "merged" in self.obj_type[self.current_obj_index]:
            size_val = self.obj_size_slider.value()
            opacity_val = self.obj_opacity_slider.value()
            compo_mode = self.obj_blend_combo.currentText()

            if self.obj_size[self.current_obj_index] != size_val:
                self.obj_size_slider.setValue(self.obj_size[self.current_obj_index])
            if self.obj_opacity[self.current_obj_index] != opacity_val:
                self.obj_opacity_slider.setValue(
                    self.obj_opacity[self.current_obj_index]
                )
            if self.obj_comp_mode[self.current_obj_index] != compo_mode:
                self.obj_blend_combo.setCurrentText(
                    self.obj_comp_mode[self.current_obj_index]
                )
        else:
            return

    def clear_all(self):
        ind = np.arange(len(self.obj_list))[::-1]
        for i in ind:
            self.layer_layout.removeWidget(self.obj_list[i])
            self.obj_list[i].deleteLater()
            del self.obj_list[i]
        self.obj_id = []
        self.obj_name = []
        self.obj_type = []
        self.obj_data = []
        self.obj_size = []
        self.obj_opacity = []
        self.obj_comp_mode = []
        self.obj_count = 0
        self.current_obj_index = None

    def compare_obj_called(self):
        compare_names = [self.obj_name[ind] for ind in self.linked_indexes]
        compare_data = [self.obj_data[ind] for ind in self.linked_indexes]
        info_window = CompareWindow(compare_names, compare_data)
        info_window.exec()



import os
import cv2
import tifffile
import colorsys
import numpy as np


# image_file_path = '/Users/jingyig/Work/Kavli/Data/HERBS_DATA/test_image_type/image_0195.tif'

class ImageReader(object):
    def __init__(self, image_file_path):
        self.error_index = 0
        self.is_czi = False
        self.file_name_list = [image_file_path[:-4]]
        if self.file_name_list[0][-1] == '.':
            self.file_name_list[0] = self.file_name_list[0][:-1]
        self.n_scenes = 1
        self.n_pages = 1

        self.scaling_val = None
        self.is_rgb = True
        self.pixel_type = 'rgb24'
        self.level = 255
        self.n_channels = 3
        self.data_type = 'uint8'
        self.rgb_colors = [(255, 0, 0), (0, 255, 0), (0, 0, 255)]
        self.channel_name = ['Red', 'Green', 'Blue']
        self.hsv_colors = []
        for i in range(3):
            chsv = colorsys.rgb_to_hsv(self.rgb_colors[i][0], self.rgb_colors[i][1], self.rgb_colors[i][2])
            hsv_color = (chsv[0], chsv[1], chsv[2] / 255)
            self.hsv_colors.append(hsv_color)
        self.gamma_val = []

        self.data = {}
        self.scale = {}

        img_data = cv2.imread(image_file_path)
        img_data = cv2.cvtColor(img_data, cv2.COLOR_BGR2RGB)

        self.data['scene 0'] = img_data
        self.scale['scene 0'] = 1



# image_file_path = '/Users/jingyig/Work/Kavli/Data/HERBS_DATA/TIF_test/13234_PHALDAB_s3_g004_tiff_30_s3.tif'
#
#
# import h5py
# filename = '/Users/jingyig/Work/Kavli/Data/HERBS_DATA/RAW-TILE_Cam_Left_00000.lux.h5'
#
# with h5py.File(filename, "r") as f:
#     # List all groups
#     all_keys = list(f.keys())
#     # Get the data
#     data = f[all_keys[0]]
#     print(data.shape)
#     plt.imshow(data[800, :, :])
#     a = data[800, :, :]
#     plt.show()
#     meta_data = f[all_keys[1]]
#
#
#
#     print(meta_data)
#
# f = h5py.File(filename, "r")
# list(f.keys())
#
# f['Data'].attrs.keys()
# f['Data'].attrs['CLASS']
# f['Data'].attrs['element_size_um']
# f.close()
# array([6.  , 1.95, 1.95])

# image_file_path = '/Users/jingyig/Work/Kavli/Data/HERBS_DATA/TIF_test/2_2_rotated.tif'


class TIFFReader(object):
    def __init__(self, image_file_path):
        self.error_index = 0
        self.is_czi = False
        self.file_name_list = [image_file_path[:-4]]
        if self.file_name_list[0][-1] == '.':
            self.file_name_list[0] = self.file_name_list[0][:-1]

        da_file = tifffile.TiffFile(image_file_path)
        self.n_scenes = len(da_file.series)

        if self.n_scenes > 1:
            self.error_index = 1
        else:
            self.n_pages = len(da_file.pages)
            self.pixel_type = da_file.pages[0].dtype
            self.software = da_file.pages[0].software
            self.is_imagej = da_file.is_imagej

            self.scaling_val = None

            self.hsv_colors = []
            self.gamma_val = []
            self.data = {}
            self.scale = {}

            if self.pixel_type == 'uint8':
                self.pixel_type = 'rgb24'
                self.level = 255
                self.data_type = 'uint8'
                self.is_rgb = True
                self.n_channels = 3
            elif self.pixel_type == 'uint16':
                self.pixel_type = 'gray16'
                self.level = 65535
                self.data_type = 'uint16'
                self.is_rgb = False
                self.n_channels = 1
            else:
                self.error_index = 2

            if da_file.is_imagej:
                if da_file.is_bigtiff:
                    if self.n_pages == 1:
                        self.error_index = 3
                    else:
                        self.data['scene 0'] = da_file.asarray()
                        self.scale['scene 0'] = 100
                        self.error_index = 7
                else:
                    if self.n_pages == 1:
                        img_data = da_file.asarray()
                        if len(img_data.shape) == 2:
                            img_data = [img_data]
                            img_data = np.dstack(img_data)
                        self.data['scene 0'] = img_data
                        self.scale['scene 0'] = 100
                    else:
                        temp = da_file.asarray()
                        img_data = []
                        for i in range(self.n_pages):
                            img_data.append(temp[i])
                        img_data = np.dstack(img_data)
                        img_data = img_data.astype(self.data_type)
                        self.data['scene 0'] = img_data
                        self.scale['scene 0'] = 100
                        self.n_channels = self.n_pages
                        self.is_rgb = False
                        self.n_pages = 1
                        self.pixel_type = 'gray8'
            else:
                if da_file.is_bigtiff:
                    if self.n_pages == 1:
                        self.error_index = 4
                    else:
                        self.data['scene 0'] = da_file.asarray()
                        self.scale['scene 0'] = 100
                        self.error_index = 7
                else:
                    if self.n_pages == 1:
                        img_data = da_file.asarray()
                        if len(img_data.shape) == 2:
                            img_data = [img_data]
                            img_data = np.dstack(img_data)
                        self.data['scene 0'] = img_data
                        self.scale['scene 0'] = 100
                    else:
                        self.error_index = 5

        if self.pixel_type == 'rgb24':
            self.rgb_colors = [(255, 0, 0), (0, 255, 0), (0, 0, 255)]
            self.channel_name = ['Red', 'Green', 'Blue']
            for i in range(3):
                chsv = colorsys.rgb_to_hsv(self.rgb_colors[i][0], self.rgb_colors[i][1], self.rgb_colors[i][2])
                hsv_color = (chsv[0], chsv[1], chsv[2] / 255)
                self.hsv_colors.append(hsv_color)
        elif self.pixel_type == 'gray16':
            self.rgb_colors = [(128, 128, 128)]
            chsv = colorsys.rgb_to_hsv(self.rgb_colors[0][0], self.rgb_colors[0][1], self.rgb_colors[0][2])
            hsv_color = (chsv[0], chsv[1], chsv[2] / 255)
            self.hsv_colors.append(hsv_color)
            self.channel_name = ['Gray']
        elif self.pixel_type == 'gray8':
            possible_rgb_color = [(128, 128, 128), (255, 0, 0), (0, 255, 0), (0, 0, 255)]
            possible_channel_name = ['Gray', 'Red', 'Green', 'Blue']
            self.rgb_colors = []
            self.channel_name = []
            for i in range(self.n_channels):
                self.rgb_colors.append(possible_rgb_color[i])
                chsv = colorsys.rgb_to_hsv(self.rgb_colors[i][0], self.rgb_colors[i][1], self.rgb_colors[i][2])
                hsv_color = (chsv[0], chsv[1], chsv[2] / 255)
                self.hsv_colors.append(hsv_color)
                self.channel_name.append(possible_channel_name[i])
        else:
            self.error_index = 6

        da_file.close()

        # for tag in da_file.pages[0].tags:
        #     tag_name, tag_value = tag.name, tag.value
        #     print(tag_name, tag_value)






class ImagesReader(object):
    def __init__(self, folder_path):

        all_files_in_folder = os.listdir(folder_path)
        self.is_czi = False
        self.is_rgb = True
        self.n_channels = 3
        self.level = 255
        self.data_type = 'uint8'
        self.hsv_colors = [(0, 255, 255), (120, 255, 255), (240, 255, 255)]
        self.rgb_colors = [(255, 0, 0), (0, 255, 0), (0, 0, 255)]
        self.channel_name = ['Red', 'Green', 'Blue']
        self.file_name_list = []
        self.data = {}
        scene_id = 0

        for i in range(len(all_files_in_folder)):
            da_image_file = all_files_in_folder[i]
            da_file_name = da_image_file[:-4]
            if da_file_name[-1] == '.':
                da_file_name = da_file_name[:-1]

            file_type = da_image_file[-4:]
            if file_type in ['.bmp', '.jpg', '.png', 'jpeg', '.tif', '.pdf']:
                self.file_name_list.append(da_file_name)
                da_file_path = os.path.join(folder_path, da_image_file)

                if file_type in ['.bmp', '.jpg', '.png', 'jpeg']:
                    data = cv2.imread(da_file_path)
                    img_data = cv2.cvtColor(data, cv2.COLOR_BGR2RGB)
                else:
                    print('TIFF file is testing')
                    img_data = tifffile.imread(da_file_path)

                self.data['scene %d' % scene_id] = img_data

                scene_id += 1

        self.n_scenes = len(self.data)



import os
import sys
import numpy as np
import pyqtgraph as pg
# from pyqtgraph.Qt import QtGui, QtCore
import pyqtgraph.functions as fn
import pyqtgraph.opengl as gl
from PyQt5.QtWidgets import *
from PyQt5.QtGui import *
from PyQt5.QtCore import *

from .image_stacks import SliceStack
from .slice_stacks import SliceStacks
from .label_tree import LabelTree
from .uuuuuu import read_qss_file, get_corner_line_from_rect, get_slice_atlas_coord, make_contour_img,  \
    rotate_base_points, rotation_x, rotation_y, rotation_z
from .probe_utiles import get_tilt_sign#, get_direction_rotation


class PageController(QWidget):
    class SignalProxy(QObject):
        sigPageChanged = pyqtSignal(object)

    def __init__(self):
        self._sigprox = PageController.SignalProxy()
        self.sig_page_changed = self._sigprox.sigPageChanged

        QWidget.__init__(self)
        page_control_style = read_qss_file('qss/page_control.qss')
        self.setStyleSheet(page_control_style)

        self.max_val = None

        self.page_slider = QSlider(Qt.Horizontal)
        self.page_slider.setMinimum(0)
        self.page_slider.valueChanged.connect(self.slider_value_changed)

        self.page_label = QLabel()
        self.page_label.setFixedSize(50, 20)

        self.page_left_btn = QPushButton()
        self.page_left_btn.setIcon(QIcon("icons/backward.svg"))
        self.page_left_btn.setIconSize(QSize(16, 16))
        self.page_left_btn.clicked.connect(self.left_btn_clicked)

        self.page_right_btn = QPushButton()
        self.page_right_btn.setIcon(QIcon("icons/forward.svg"))
        self.page_right_btn.setIconSize(QSize(16, 16))
        self.page_right_btn.clicked.connect(self.right_btn_clicked)

        self.page_fast_left_btn = QPushButton()
        self.page_fast_left_btn.setIcon(QIcon("icons/fast_backward.svg"))
        self.page_fast_left_btn.setIconSize(QSize(16, 16))
        self.page_fast_left_btn.clicked.connect(self.fast_left_btn_clicked)

        self.page_fast_right_btn = QPushButton()
        self.page_fast_right_btn.setIcon(QIcon("icons/fast_forward.svg"))
        self.page_fast_right_btn.setIconSize(QSize(16, 16))
        self.page_fast_right_btn.clicked.connect(self.fast_right_btn_clicked)

        page_ctrl_layout = QHBoxLayout()
        page_ctrl_layout.setSpacing(0)
        page_ctrl_layout.setContentsMargins(10, 5, 10, 5)
        page_ctrl_layout.addWidget(self.page_left_btn)
        page_ctrl_layout.addSpacing(10)
        page_ctrl_layout.addWidget(self.page_fast_left_btn)
        page_ctrl_layout.addSpacing(10)
        page_ctrl_layout.addWidget(self.page_slider)
        page_ctrl_layout.addSpacing(5)
        page_ctrl_layout.addWidget(self.page_label)
        page_ctrl_layout.addWidget(self.page_fast_right_btn)
        page_ctrl_layout.addSpacing(10)
        page_ctrl_layout.addWidget(self.page_right_btn)

        self.setLayout(page_ctrl_layout)

    def set_val(self, val):
        self.page_slider.setValue(val)

    def set_max(self, val):
        self.max_val = val
        self.page_slider.setMaximum(val)

    def slider_value_changed(self):
        val = self.page_slider.value()
        self.page_label.setText(str(val))
        self.sig_page_changed.emit(val)

    def left_btn_clicked(self):
        if self.max_val is None:
            return
        val = self.page_slider.value() - 1
        if val < 0:
            val = 0
        self.set_val(val)

    def right_btn_clicked(self):
        if self.max_val is None:
            return
        val = self.page_slider.value() + 1
        if val > self.max_val:
            val = self.max_val
        self.set_val(val)

    def fast_left_btn_clicked(self):
        if self.max_val is None:
            return
        val = self.page_slider.value() - 10
        if val < 0:
            val = 0
        self.set_val(val)

    def fast_right_btn_clicked(self):
        if self.max_val is None:
            return
        val = self.page_slider.value() + 10
        if val > self.max_val:
            val = self.max_val
        self.set_val(val)


class SliceRotation(QWidget):

    class SignalProxy(QObject):
        sigSliceRotated = pyqtSignal(object)  # id

    def __init__(self):
        self._sigprox = SliceRotation.SignalProxy()
        self.sig_slice_rotated = self._sigprox.sigSliceRotated

        QWidget.__init__(self)

        spinbox_line_edit_style = read_qss_file('qss/hidden_line_edit.qss')
        self.prec = 0.01
        self.rot_range = 30
        self.n_steps = int(self.rot_range / self.prec)

        self.h_slider = QSlider(Qt.Horizontal)
        self.h_slider.valueChanged.connect(self.h_slider_changed)
        self.h_slider.setValue(0)
        self.h_slider.setRange(-self.n_steps, self.n_steps)

        self.h_spinbox = QDoubleSpinBox()
        self.h_spinbox.lineEdit().setStyleSheet(spinbox_line_edit_style)
        self.h_spinbox.valueChanged.connect(self.h_spinbox_changed)
        self.h_spinbox.setDecimals(2)
        self.h_spinbox.setValue(0)
        self.h_spinbox.setRange(-self.rot_range, self.rot_range)
        self.h_spinbox.setSingleStep(self.prec)
        self.h_spinbox.setMinimumSize(50, 20)

        self.v_slider = QSlider(Qt.Horizontal)
        self.v_slider.valueChanged.connect(self.v_slider_changed)
        self.v_slider.setValue(0)
        self.v_slider.setRange(-self.n_steps, self.n_steps)

        self.v_spinbox = QDoubleSpinBox()
        self.v_spinbox.lineEdit().setStyleSheet(spinbox_line_edit_style)
        self.v_spinbox.valueChanged.connect(self.v_spinbox_changed)
        self.v_spinbox.setDecimals(2)
        self.v_spinbox.setValue(0)
        self.v_spinbox.setRange(-self.rot_range, self.rot_range)
        self.v_spinbox.setSingleStep(self.prec)
        self.v_spinbox.setMinimumSize(50, 20)

    def set_prec(self, prec):
        self.prec = prec

    def h_slider_changed(self):
        val = self.h_slider.value() * self.prec
        self.h_spinbox.setValue(val)

    def h_spinbox_changed(self):
        val = int(self.h_spinbox.value() / self.prec)
        if self.h_slider.value() != val:
            self.h_slider.setValue(val)
        self.sig_slice_rotated.emit(np.deg2rad([self.h_spinbox.value(), self.v_spinbox.value()]))

    def v_slider_changed(self):
        val = self.v_slider.value() * self.prec
        self.v_spinbox.setValue(val)

    def v_spinbox_changed(self):
        val = int(self.v_spinbox.value() / self.prec)
        if self.v_slider.value() != val:
            self.v_slider.setValue(val)
        self.sig_slice_rotated.emit(np.deg2rad([self.h_spinbox.value(), self.v_spinbox.value()]))


class ImageLabel(QWidget):
    def __init__(self, img, title):
        super(QWidget, self).__init__()
        layout = QHBoxLayout(self)
        self.label3 = QLabel(self)
        self.title = QLabel(title)
        self.pixmap = QPixmap(img)
        self.pixmap = self.pixmap.scaled(QSize(20, 10))
        self.label3.setPixmap(self.pixmap)
        self.label3.setAlignment(Qt.AlignCenter)
        self.title.setMinimumHeight(self.pixmap.height())
        self.title.setAlignment(Qt.AlignCenter)
        layout.addWidget(self.label3)
        layout.addWidget(self.title)
        self.label3.setStyleSheet('background-color: transparent;')
        self.title.setStyleSheet("""
            background-color: transparent;
            color: black;
            padding: 0px 3px 0px 0px;
        """)
        layout.setSpacing(0)
        layout.addStretch()


class RotationBtn(QPushButton):
    def __init__(self, icon_path):
        super(QPushButton, self).__init__()
        btn_style = read_qss_file('qss/rotation_button.qss')
        self.setStyleSheet(btn_style)
        self.setFixedSize(15, 15)
        self.setIcon(QIcon(icon_path))
        self.setIconSize(QSize(15, 15))


class AtlasView(QObject):
    """
    A collection of user interface elements bound together:


    """
    def __init__(self):
        super(AtlasView, self).__init__()
        QObject.__init__(self)

        self.atlas_data = None
        self.atlas_label = None
        self.label_info = None
        self.atlas_boundary = None
        self.slice_size = None
        self.slice_tb_size_base = 80
        self.coronal_tb_size = None
        self.sagital_tb_size = None
        self.horizontal_tb_size = None
        self.slice_tb_size = None

        self.slice_image_data = None
        self.processing_slice = None
        self.slice_cut = 'Coronal'
        self.slice_width = 0
        self.slice_height = 0
        self.slice_distance = 0
        self.slice_bregma = []
        self.slice_boundary_points = []
        self.slice_info_ready = False

        self.remain_angle = False
        self.coronal_rotated = False
        self.sagittal_rotated = False
        self.horizontal_rotated = False
        self.coronal_rads = (0, 0)
        self.sagittal_rads = (0, 0)
        self.horizontal_rads = (0, 0)

        self.atlas_size = None
        self.anterior_info = None
        self.dorsal_info = None
        self.right_info = None
        self.vox_size_um = None
        self.Bregma = None
        self.rotate_origin_3d = None
        self.origin_3d = None

        self.c_size = None
        self.s_size = None
        self.h_size = None

        self.c_rotm_2d = None
        self.s_rotm_2d = None
        self.h_rotm_2d = None

        self.c_rotm_3d = None
        self.s_rotm_3d = None
        self.h_rotm_3d = None

        self.side_lines = None
        self.corner_points = None

        self.current_coronal_index = None
        self.current_sagital_index = None
        self.current_horizontal_index = None

        self.anchor_coronal_index = None
        self.anchor_sagital_index = None
        self.anchor_horizontal_index = None

        self.has_display_objects = False

        rotation_gb_style = read_qss_file('qss/atlas_view_group_box.qss')
        button_style = read_qss_file('qss/side_bar.qss')
        spinbox_line_edit_style = read_qss_file('qss/hidden_line_edit.qss')

        # self.display_atlas = None
        # self.display_label = None
        # self.scale = None
        # self.interpolate = True
        # self.oy_vector = np.array([0, 1, 0])
        # self.oz_vector = np.array([0, 0, 1])
        # self.ox_vector = np.array([1, 0, 0])

        self.axis = gl.GLAxisItem()
        self.axis.setSize(250, 700, 250)

        self.grid = gl.GLGridItem()
        self.grid.scale(2, 2, 1)

        self.mesh = gl.GLMeshItem(smooth=True, color=[0.5, 0.5, 0.5, 0.2], shader='balloon')
        self.mesh.setGLOptions('additive')

        # 3d things
        plate_color = [0.5, 0.5, 0.5, 0.2]
        self.ap_plate_verts = np.array([[-1, 0, -1], [-1, 0, 1], [1, 0, 1], [1, 0, -1]])
        self.ap_plate_faces = np.array([[0, 1, 2], [0, 2, 3]])
        ap_plate_md = gl.MeshData(vertexes=self.ap_plate_verts, faces=self.ap_plate_faces)
        self.ap_plate_mesh = gl.GLMeshItem(meshdata=ap_plate_md, smooth=False, color=plate_color)
        self.ap_plate_mesh.setGLOptions('additive')

        self.dv_plate_verts = np.array([[-1, -1, 0], [-1, 1, 0], [1, 1, 0], [1, -1, 0]])
        self.dv_plate_faces = np.array([[0, 1, 2], [0, 2, 3]])
        dv_plate_md = gl.MeshData(vertexes=self.dv_plate_verts, faces=self.dv_plate_faces)
        self.dv_plate_mesh = gl.GLMeshItem(meshdata=dv_plate_md, smooth=False, color=plate_color)
        self.dv_plate_mesh.setGLOptions('additive')

        self.ml_plate_verts = np.array([[0, -1, -1], [0, 1, -1], [0, 1, 1], [0, -1, 1]])
        self.ml_plate_faces = np.array([[0, 1, 2], [0, 2, 3]])
        ml_plate_md = gl.MeshData(vertexes=self.ml_plate_verts, faces=self.ml_plate_faces)
        self.ml_plate_mesh = gl.GLMeshItem(meshdata=ml_plate_md, smooth=False, color=plate_color)
        self.ml_plate_mesh.setGLOptions('additive')

        # ap - coronal
        self.cimg = SliceStacks()
        self.cpage_ctrl = PageController()
        self.cpage_ctrl.sig_page_changed.connect(self.coronal_slice_page_changed)
        self.clut = pg.HistogramLUTWidget()
        self.clut.setImageItem(self.cimg.img)

        self.simg = SliceStacks()
        self.spage_ctrl = PageController()
        self.spage_ctrl.sig_page_changed.connect(self.sagital_slice_page_changed)
        self.slut = pg.HistogramLUTWidget()
        self.slut.setImageItem(self.simg.img)

        self.himg = SliceStacks()
        self.hpage_ctrl = PageController()
        self.hpage_ctrl.sig_page_changed.connect(self.horizontal_slice_page_changed)
        self.hlut = pg.HistogramLUTWidget()
        self.hlut.setImageItem(self.himg.img)

        self.label_tree = LabelTree()

        self.working_atlas = self.cimg
        self.working_page_control = self.cpage_ctrl

        self.slice_stack = SliceStack()


        # radio buttons
        self.radio_group = QFrame()
        self.radio_group.setFixedHeight(50)
        self.radio_group.setStyleSheet('QFrame{border: 1px solid gray; border-radius: 3px}')
        radio_group_layout = QHBoxLayout(self.radio_group)
        radio_group_layout.setContentsMargins(5, 0, 5, 0)
        radio_group_layout.setAlignment(Qt.AlignCenter)
        self.section_rabnt1 = QRadioButton('Coronal')
        self.section_rabnt1.setChecked(True)
        self.section_rabnt2 = QRadioButton('Sagittal')
        self.section_rabnt3 = QRadioButton('Horizontal')
        self.section_rabnt1.setStyleSheet('color: white')
        self.section_rabnt2.setStyleSheet('color: white')
        self.section_rabnt3.setStyleSheet('color: white')
        radio_group_layout.addWidget(self.section_rabnt1)
        radio_group_layout.addWidget(self.section_rabnt2)
        radio_group_layout.addWidget(self.section_rabnt3)

        # angles
        self.crotation_ctrl = SliceRotation()
        self.crotation_ctrl.sig_slice_rotated.connect(self.coronal_slice_rotated)
        self.srotation_ctrl = SliceRotation()
        self.srotation_ctrl.sig_slice_rotated.connect(self.sagital_slice_rotated)
        self.hrotation_ctrl = SliceRotation()
        self.hrotation_ctrl.sig_slice_rotated.connect(self.horizontal_slice_rotated)


        # opacity
        atlas_op_label = QLabel('Opacity: ')
        self.atlas_op_slider = QSlider(Qt.Horizontal)
        self.atlas_op_slider.setValue(100)
        self.atlas_op_slider.setMinimum(0)
        self.atlas_op_slider.setMaximum(100)
        self.atlas_op_slider.valueChanged.connect(self.change_opacity_spinbox_value)
        # self.atlas_op_slider.valueChanged.connect(self.sig_rescale_slider)

        self.atlas_op_spinbox = QDoubleSpinBox()
        self.atlas_op_spinbox.lineEdit().setStyleSheet(spinbox_line_edit_style)
        self.atlas_op_spinbox.setValue(1)
        self.atlas_op_spinbox.setRange(0, 1)
        self.atlas_op_spinbox.setSingleStep(0.05)
        self.atlas_op_spinbox.setMinimumSize(50, 20)
        self.atlas_op_spinbox.valueChanged.connect(self.opacity_changed)

        opacity_wrap = QFrame()
        opacity_layout = QHBoxLayout(opacity_wrap)
        opacity_layout.setSpacing(2)
        opacity_layout.setContentsMargins(0, 0, 0, 0)
        opacity_layout.addWidget(atlas_op_label)
        opacity_layout.addWidget(self.atlas_op_slider)
        opacity_layout.addWidget(self.atlas_op_spinbox)

        # show brain region boundary of a volume atlas slice
        self.show_boundary_btn = QPushButton('Show Boundary')
        self.show_boundary_btn.setStyleSheet(button_style)
        self.show_boundary_btn.setCheckable(True)

        # turn on/off navigation
        self.navigation_btn = QPushButton('Navigation')
        self.navigation_btn.setStyleSheet(button_style)
        self.navigation_btn.setCheckable(True)
        self.navigation_btn.clicked.connect(self.navigation_btn_clicked)

        # turn on/off rotation remain
        self.rotation_remain_btn = QPushButton('Keep Slice Angles')
        self.rotation_remain_btn.setStyleSheet(button_style)
        self.rotation_remain_btn.setCheckable(True)
        self.rotation_remain_btn.clicked.connect(self.rotation_remain_btn_clicked)

        # coronal section control
        coronal_rotation_wrap = QGroupBox('Coronal Section')
        coronal_rotation_wrap.setStyleSheet(rotation_gb_style)
        coronal_wrap_layout = QGridLayout(coronal_rotation_wrap)
        coronal_wrap_layout.setContentsMargins(0, 0, 0, 0)
        coronal_wrap_layout.setSpacing(10)

        c_section_label_img = QPixmap('icons/sidebar/c_section.png')
        c_section_label_img = c_section_label_img.scaled(QSize(40, 40))
        c_section_label = QLabel('Tilt Z: ')
        c_section_label.setPixmap(c_section_label_img)

        c_lr_btn = RotationBtn("icons/sidebar/rotation_horizontal.svg")
        c_ud_btn = RotationBtn("icons/sidebar/rotation_vertical.svg")

        coronal_wrap_layout.addWidget(c_section_label, 0, 0, 2, 1)
        coronal_wrap_layout.addWidget(c_lr_btn, 0, 1, 1, 1)
        coronal_wrap_layout.addWidget(self.crotation_ctrl.h_slider, 0, 2, 1, 1)
        coronal_wrap_layout.addWidget(self.crotation_ctrl.h_spinbox, 0, 3, 1, 1)
        coronal_wrap_layout.addWidget(c_ud_btn, 1, 1, 1, 1)
        coronal_wrap_layout.addWidget(self.crotation_ctrl.v_slider, 1, 2, 1, 1)
        coronal_wrap_layout.addWidget(self.crotation_ctrl.v_spinbox, 1, 3, 1, 1)

        # sagital section control
        sagital_rotation_wrap = QGroupBox('Sagittal Section')
        sagital_rotation_wrap.setStyleSheet(rotation_gb_style)
        sagital_wrap_layout = QGridLayout(sagital_rotation_wrap)
        sagital_wrap_layout.setContentsMargins(0, 0, 0, 0)
        sagital_wrap_layout.setSpacing(10)

        s_section_label_img = QPixmap('icons/sidebar/s_section.png')
        s_section_label_img = s_section_label_img.scaled(QSize(40, 40))
        s_section_label = QLabel()
        s_section_label.setPixmap(s_section_label_img)

        s_lr_btn = RotationBtn("icons/sidebar/rotation_horizontal.svg")
        s_ud_btn = RotationBtn("icons/sidebar/rotation_vertical.svg")

        sagital_wrap_layout.addWidget(s_section_label, 0, 0, 2, 1)
        sagital_wrap_layout.addWidget(s_lr_btn, 0, 1, 1, 1)
        sagital_wrap_layout.addWidget(self.srotation_ctrl.h_slider, 0, 2, 1, 1)
        sagital_wrap_layout.addWidget(self.srotation_ctrl.h_spinbox, 0, 3, 1, 1)
        sagital_wrap_layout.addWidget(s_ud_btn, 1, 1, 1, 1)
        sagital_wrap_layout.addWidget(self.srotation_ctrl.v_slider, 1, 2, 1, 1)
        sagital_wrap_layout.addWidget(self.srotation_ctrl.v_spinbox, 1, 3, 1, 1)

        # horizontal section control
        horizontal_rotation_wrap = QGroupBox('Horizontal Section')
        horizontal_rotation_wrap.setStyleSheet(rotation_gb_style)
        horizontal_wrap_layout = QGridLayout(horizontal_rotation_wrap)
        horizontal_wrap_layout.setContentsMargins(0, 0, 0, 0)
        horizontal_wrap_layout.setSpacing(10)

        h_section_label_img = QPixmap('icons/sidebar/h_section.png')
        h_section_label_img = h_section_label_img.scaled(QSize(40, 40))
        h_section_label = QLabel()
        h_section_label.setPixmap(h_section_label_img)

        h_lr_btn = RotationBtn("icons/sidebar/rotation_horizontal.svg")
        h_ud_btn = RotationBtn("icons/sidebar/rotation_vertical.svg")

        horizontal_wrap_layout.addWidget(h_section_label, 0, 0, 2, 1)
        horizontal_wrap_layout.addWidget(h_lr_btn, 0, 1, 1, 1)
        horizontal_wrap_layout.addWidget(self.hrotation_ctrl.h_slider, 0, 2, 1, 1)
        horizontal_wrap_layout.addWidget(self.hrotation_ctrl.h_spinbox, 0, 3, 1, 1)
        horizontal_wrap_layout.addWidget(h_ud_btn, 1, 1, 1, 1)
        horizontal_wrap_layout.addWidget(self.hrotation_ctrl.v_slider, 1, 2, 1, 1)
        horizontal_wrap_layout.addWidget(self.hrotation_ctrl.v_spinbox, 1, 3, 1, 1)

        self.sidebar_wrap = QFrame()
        sidebar_wrap_layout = QVBoxLayout(self.sidebar_wrap)
        sidebar_wrap_layout.setSpacing(0)
        sidebar_wrap_layout.addWidget(self.radio_group)
        sidebar_wrap_layout.addSpacing(10)
        sidebar_wrap_layout.addWidget(opacity_wrap)
        sidebar_wrap_layout.addSpacing(10)
        sidebar_wrap_layout.addWidget(self.show_boundary_btn)
        sidebar_wrap_layout.addSpacing(10)
        sidebar_wrap_layout.addWidget(coronal_rotation_wrap)
        sidebar_wrap_layout.addSpacing(10)
        sidebar_wrap_layout.addWidget(sagital_rotation_wrap)
        sidebar_wrap_layout.addSpacing(10)
        sidebar_wrap_layout.addWidget(horizontal_rotation_wrap)
        sidebar_wrap_layout.addSpacing(5)
        sidebar_wrap_layout.addWidget(self.rotation_remain_btn)
        sidebar_wrap_layout.addStretch(1)
        sidebar_wrap_layout.addWidget(self.navigation_btn)

        # self.lut.sigLookupTableChanged.connect(self.histlut_changed)
        # self.lut.sigLevelsChanged.connect(self.histlut_changed)

    def reset_3d_plates(self):
        self.ap_plate_verts = np.array([[-1, 0, -1], [-1, 0, 1], [1, 0, 1], [1, 0, -1]])
        self.ap_plate_faces = np.array([[0, 1, 2], [0, 2, 3]])

        self.dv_plate_verts = np.array([[-1, -1, 0], [-1, 1, 0], [1, 1, 0], [1, -1, 0]])
        self.dv_plate_faces = np.array([[0, 1, 2], [0, 2, 3]])

        self.ml_plate_verts = np.array([[0, -1, -1], [0, 1, -1], [0, 1, 1], [0, -1, 1]])
        self.ml_plate_faces = np.array([[0, 1, 2], [0, 2, 3]])

    def set_slice_data(self, slice_data):
        self.slice_image_data = slice_data.copy()
        self.slice_stack.set_data(slice_data)
        self.slice_size = np.ravel(self.slice_stack.img_layer.image.shape[:2])
        slice_scale = np.max(self.slice_size / self.slice_tb_size_base)
        self.slice_tb_size = self.slice_size / slice_scale
        self.slice_tb_size = self.slice_tb_size[::-1].astype(int)

        self.working_atlas = self.slice_stack

        rect = (0, 0, int(self.slice_size[1]), int(self.slice_size[0]))
        self.corner_points, self.side_lines = get_corner_line_from_rect(rect)
        self.working_atlas.image_dict['tri_pnts'].set_range(x_range=self.slice_size[1] - 1,
                                                            y_range=self.slice_size[0] - 1)

    def set_slice_info(self, slice_info):
        self.slice_cut = slice_info.cut
        self.slice_width = slice_info.width
        self.slice_height = slice_info.height
        self.slice_distance = slice_info.distance
        if self.slice_height == 0 or self.slice_width == 0 or self.slice_distance == 0 or not self.slice_bregma:
            self.slice_info_ready = False
        else:
            self.slice_info_ready = True

    def save_slice_data_and_info(self):
        data = {'data': self.slice_image_data,
                'cut': self.slice_cut,
                'width': self.slice_width,
                'height': self.slice_height,
                'distance': self.slice_distance,
                'Bregma': self.slice_bregma,
                'ready': self.slice_info_ready}
        return data

    def set_slice_data_and_info(self, slice_data):
        img_data = slice_data['data'].copy()
        self.slice_cut = slice_data['cut']
        self.slice_width = slice_data['width']
        self.slice_height = slice_data['height']
        self.slice_distance = slice_data['distance']
        self.slice_bregma = slice_data['Bregma']
        self.slice_info_ready = slice_data['ready']
        self.set_slice_data(img_data)

    def get_slice_coords(self, ruler_data):
        x, y, z = get_slice_atlas_coord(ruler_data, self.slice_cut, self.slice_size,
                                        self.slice_width, self.slice_height, self.slice_distance, self.slice_bregma)
        return x, y, z

    def check_info_ready(self):
        if self.slice_height == 0 or self.slice_width == 0 or self.slice_distance == 0:
            self.slice_info_ready = False
        else:
            self.slice_info_ready = True

    def clear_slice_info(self):
        if self.slice_image_data is not None:
            self.slice_cut = 'Coronal'
            self.slice_width = 0
            self.slice_height = 0
            self.slice_distance = 0
            self.slice_bregma = []
            self.slice_info_ready = False
        else:
            return

    def set_data(self, atlas_data, atlas_label, atlas_info, label_info, boundaries):
        if self.atlas_data is not None:
            self.reset_3d_plates()
            self.clear_all_display_obj()

        self.atlas_data = atlas_data
        self.atlas_label = atlas_label
        self.label_info = label_info
        self.atlas_boundary = boundaries
        self.label_tree.set_labels(label_info)

        self.atlas_size = self.atlas_data.shape
        self.anterior_info = atlas_info[0]
        self.dorsal_info = atlas_info[1]
        self.right_info = atlas_info[2]
        self.vox_size_um = atlas_info[3]['vxsize']

        if np.all(np.ravel(atlas_info[3]['Bregma']) == 0):
            da_bregma = (int(self.atlas_size[1] / 2), int(self.atlas_size[2] / 2), int(self.atlas_size[0] / 2))
        else:
            da_bregma = atlas_info[3]['Bregma']
        inverse_b2 = self.atlas_size[0] - 1 - da_bregma[2]
        self.Bregma = (inverse_b2, da_bregma[0], da_bregma[1])
        print(self.Bregma, 'bregma')
        self.rotate_origin_3d = np.array(da_bregma)
        self.origin_3d = np.array(da_bregma)

        self.current_coronal_index = self.origin_3d[1]
        self.current_sagital_index = self.origin_3d[0]
        self.current_horizontal_index = self.origin_3d[2]

        self.cpage_ctrl.set_max(self.atlas_size[2] - 1)
        self.spage_ctrl.set_max(self.atlas_size[1] - 1)
        self.hpage_ctrl.set_max(self.atlas_size[0] - 1)

        if self.current_coronal_index == 0:
            self.coronal_slice_page_changed(self.current_coronal_index)
            self.cpage_ctrl.page_label.setText('0')
        else:
            self.cpage_ctrl.set_val(self.current_coronal_index)

        if self.current_sagital_index == 0:
            self.sagital_slice_page_changed(self.current_sagital_index)
            self.spage_ctrl.page_label.setText('0')
        else:
            self.spage_ctrl.set_val(self.current_sagital_index)

        if self.current_horizontal_index == 0:
            self.horizontal_slice_page_changed(self.current_horizontal_index)
            self.hpage_ctrl.page_label.setText('0')
        else:
            self.hpage_ctrl.set_val(self.current_horizontal_index)

        self.cimg.label_img.setLevels(levels=[0, self.label_tree.label_level])
        self.simg.label_img.setLevels(levels=[0, self.label_tree.label_level])
        self.himg.label_img.setLevels(levels=[0, self.label_tree.label_level])

        lut = self.label_tree.lookup_table()
        self.cimg.label_img.setLookupTable(lut=lut)
        self.simg.label_img.setLookupTable(lut=lut)
        self.himg.label_img.setLookupTable(lut=lut)

        self.c_size = np.array([self.atlas_size[0], self.atlas_size[1]])
        coronal_scale = np.max(self.c_size / self.slice_tb_size_base)
        self.coronal_tb_size = self.c_size[::-1] / coronal_scale
        self.coronal_tb_size = self.coronal_tb_size.astype(int)

        self.s_size = np.array([self.atlas_size[0], self.atlas_size[2]])
        sagital_scale = np.max(self.s_size / self.slice_tb_size_base)
        self.sagital_tb_size = self.s_size[::-1] / sagital_scale
        self.sagital_tb_size = self.sagital_tb_size.astype(int)

        self.h_size = np.array([self.atlas_size[1], self.atlas_size[2]])
        horizontal_scale = np.max(self.h_size / self.slice_tb_size_base)
        self.horizontal_tb_size = self.h_size[::-1] / horizontal_scale
        self.horizontal_tb_size = self.horizontal_tb_size.astype(int)

        self.update_3d_plate_component()

    def update_3d_plate_component(self):
        self.ap_plate_verts[:, 0] = self.ap_plate_verts[:, 0] * self.atlas_size[1] * 0.5
        self.ap_plate_verts[:, 2] = self.ap_plate_verts[:, 2] * self.atlas_size[0] * 0.5

        self.dv_plate_verts[:, 0] = self.dv_plate_verts[:, 0] * self.atlas_size[1] * 0.5
        self.dv_plate_verts[:, 1] = self.dv_plate_verts[:, 1] * self.atlas_size[2] * 0.5

        self.ml_plate_verts[:, 1] = self.ml_plate_verts[:, 1] * self.atlas_size[2] * 0.5
        self.ml_plate_verts[:, 2] = self.ml_plate_verts[:, 2] * self.atlas_size[0] * 0.5

        z_slide_dist = (self.atlas_size[0] * 0.5 - self.origin_3d[2])
        y_slide_dist = (self.atlas_size[2] * 0.5 - self.origin_3d[1])
        x_slide_dist = (self.atlas_size[1] * 0.5 - self.origin_3d[0])

        self.ap_plate_verts = self.ap_plate_verts + np.array([x_slide_dist, 0, z_slide_dist])
        self.dv_plate_verts = self.dv_plate_verts + np.array([x_slide_dist, y_slide_dist, 0])
        self.ml_plate_verts = self.ml_plate_verts + np.array([0, y_slide_dist, z_slide_dist])

        ap_plate_md = gl.MeshData(vertexes=self.ap_plate_verts, faces=self.ap_plate_faces)
        self.ap_plate_mesh.setMeshData(meshdata=ap_plate_md)

        dv_plate_md = gl.MeshData(vertexes=self.dv_plate_verts, faces=self.dv_plate_faces)
        self.dv_plate_mesh.setMeshData(meshdata=dv_plate_md)

        ml_plate_md = gl.MeshData(vertexes=self.ml_plate_verts, faces=self.ml_plate_faces)
        self.ml_plate_mesh.setMeshData(meshdata=ml_plate_md)

    def working_cut_changed(self, atlas_display):
        self.pre_trajectory_changed()
        if atlas_display == "coronal":
            self.working_atlas = self.cimg
            self.working_page_control = self.cpage_ctrl
            self.slice_tb_size = self.coronal_tb_size
            self.slice_size = self.c_size
        elif atlas_display == "sagittal":
            self.working_atlas = self.simg
            self.working_page_control = self.spage_ctrl
            self.slice_tb_size = self.sagital_tb_size
            self.slice_size = self.s_size
        else:
            self.working_atlas = self.himg
            self.working_page_control = self.hpage_ctrl
            self.slice_tb_size = self.horizontal_tb_size
            self.slice_size = self.h_size
        rect = (0, 0, int(self.slice_size[1]), int(self.slice_size[0]))
        self.corner_points, self.side_lines = get_corner_line_from_rect(rect)
        self.working_atlas.image_dict['tri_pnts'].set_range(x_range=self.slice_size[1] - 1,
                                                            y_range=self.slice_size[0] - 1)

    def navigation_btn_clicked(self):
        if self.navigation_btn.isChecked():
            # self.
            self.cimg.v_line.setVisible(True)
            self.cimg.h_line.setVisible(True)
            self.himg.v_line.setVisible(True)
            self.himg.h_line.setVisible(True)
            self.simg.v_line.setVisible(True)
            self.simg.h_line.setVisible(True)
        else:
            self.cimg.v_line.setVisible(False)
            self.cimg.h_line.setVisible(False)
            self.himg.v_line.setVisible(False)
            self.himg.h_line.setVisible(False)
            self.simg.v_line.setVisible(False)
            self.simg.h_line.setVisible(False)

    def rotation_remain_btn_clicked(self):
        if self.rotation_remain_btn.isChecked():
            self.remain_angle = True
        else:
            self.remain_angle = False

    def change_opacity_spinbox_value(self):
        val = self.atlas_op_slider.value()
        self.atlas_op_spinbox.setValue(val / 100)

    def opacity_changed(self):
        val = self.atlas_op_spinbox.value()
        if self.atlas_op_slider.value() != int(val * 100):
            self.atlas_op_slider.setValue(int(val * 100))
        self.cimg.label_img.setOpts(opacity=val)
        self.simg.label_img.setOpts(opacity=val)
        self.himg.label_img.setOpts(opacity=val)

    # coronal page number changed and slice rotated
    def set_perfect_coronal_slice(self):
        da_atlas_slice = self.atlas_data[:, :, self.current_coronal_index]
        da_atlas_label = self.atlas_label[:, :, self.current_coronal_index]
        da_atlas_contour = self.atlas_boundary['c_contour'][:, :, self.current_coronal_index]
        self.cimg.set_data(da_atlas_slice, da_atlas_label, da_atlas_contour, scale=None)

        slide_dist = (self.current_coronal_index - self.origin_3d[1])
        ap_plate_verts = self.ap_plate_verts + np.array([0, slide_dist, 0])
        ap_plate_md = gl.MeshData(vertexes=ap_plate_verts, faces=self.ap_plate_faces)
        self.ap_plate_mesh.setMeshData(meshdata=ap_plate_md)
        self.get_3d_origin()

    def rotate_coronal_current_slice(self):
        # calculate for 2d rotation
        c_id = self.current_coronal_index
        s_id = self.current_sagital_index
        h_id = self.atlas_size[0] - 1 - self.current_horizontal_index

        z_angle = self.coronal_rads[0]
        x_angle = self.coronal_rads[1]
        self.c_rotm_2d = np.dot(rotation_x(z_angle), rotation_y(x_angle))

        o_val = np.array([0, 0, c_id])
        o_rot = np.array([h_id, s_id, c_id])

        oz_vector = np.dot(self.c_rotm_2d, np.array([1, 0, 0]))
        ox_vector = np.dot(self.c_rotm_2d, np.array([0, 1, 0]))

        oval_new = o_rot + np.dot(self.c_rotm_2d, o_val - o_rot)

        ox_length = self.atlas_size[1]
        oz_length = self.atlas_size[0]

        atlas = fn.affineSlice(self.atlas_data, shape=(oz_length, ox_length), vectors=[oz_vector, ox_vector],
                               origin=(oval_new[0], oval_new[1], oval_new[2]), axes=(0, 1, 2), order=1)
        label = fn.affineSlice(self.atlas_label, shape=(oz_length, ox_length), vectors=[oz_vector, ox_vector],
                               origin=(oval_new[0], oval_new[1], oval_new[2]), axes=(0, 1, 2), order=0)
        da_contour = make_contour_img(label)
        self.cimg.set_data(atlas, label, da_contour, scale=None)

        self.c_rotm_3d = np.dot(rotation_z(z_angle), rotation_x(-x_angle))

        slide_dist = self.current_coronal_index - self.origin_3d[1]
        ap_plate_verts = self.ap_plate_verts + np.array([0, slide_dist, 0])
        ap_plate_verts = np.dot(self.c_rotm_3d, (ap_plate_verts - self.rotate_origin_3d).T).T + self.rotate_origin_3d
        ap_plate_md = gl.MeshData(vertexes=ap_plate_verts, faces=self.ap_plate_faces)
        self.ap_plate_mesh.setMeshData(meshdata=ap_plate_md)

    def coronal_slice_page_changed(self, page_number):
        self.clear_all_display_obj()
        if self.atlas_data is None or self.atlas_label is None:
            return
        self.current_coronal_index = page_number
        self.set_perfect_coronal_slice()
        if self.remain_angle:
            if self.coronal_rotated:
                self.rotate_coronal_current_slice()
        else:
            if self.coronal_rotated:
                self.crotation_ctrl.h_spinbox.blockSignals(True)
                self.crotation_ctrl.v_spinbox.blockSignals(True)
                self.crotation_ctrl.h_slider.setValue(0)
                self.crotation_ctrl.v_slider.setValue(0)
                self.crotation_ctrl.h_spinbox.blockSignals(False)
                self.crotation_ctrl.v_spinbox.blockSignals(False)
                self.coronal_rotated = False

    def coronal_slice_rotated(self, rads):
        self.clear_all_display_obj()
        if self.atlas_data is None or self.atlas_label is None:
            return
        self.coronal_rads = rads
        if np.all(np.ravel(rads) == 0):
            if self.coronal_rotated:
                self.set_perfect_coronal_slice()
                self.coronal_rotated = False
        else:
            self.rotate_coronal_current_slice()
            self.coronal_rotated = True

    # sagittal page number changed and slice rotated
    def set_perfect_sagittal_slice(self):
        da_atlas_slice = self.atlas_data[:, self.current_sagital_index, :]
        da_atlas_label = self.atlas_label[:, self.current_sagital_index, :]
        da_atlas_contour = self.atlas_boundary['s_contour'][:, self.current_sagital_index, :]
        self.simg.set_data(da_atlas_slice, da_atlas_label, da_atlas_contour, scale=None)

        slide_dist = (self.current_sagital_index - self.origin_3d[0])
        ml_plate_verts = self.ml_plate_verts + np.array([slide_dist, 0, 0])
        ml_plate_md = gl.MeshData(vertexes=ml_plate_verts, faces=self.ml_plate_faces)
        self.ml_plate_mesh.setMeshData(meshdata=ml_plate_md)
        self.get_3d_origin()

    def rotate_sagittal_current_slice(self):
        # calculate for 2d rotation
        c_id = self.current_coronal_index
        s_id = self.current_sagital_index
        h_id = self.atlas_size[0] - 1 - self.current_horizontal_index

        z_angle = self.sagittal_rads[0]
        y_angle = self.sagittal_rads[1]
        self.s_rotm_2d = np.dot(rotation_x(z_angle), rotation_z(y_angle))

        o_val = np.array([0, s_id, 0])
        o_rot = np.array([h_id, s_id, c_id])

        oz_vector = np.dot(self.s_rotm_2d, np.array([1, 0, 0]))
        oy_vector = np.dot(self.s_rotm_2d, np.array([0, 0, 1]))

        oval_new = o_rot + np.dot(self.s_rotm_2d, o_val - o_rot)

        oy_length = self.atlas_size[2]
        oz_length = self.atlas_size[0]

        atlas = fn.affineSlice(self.atlas_data, shape=(oz_length, oy_length), vectors=[oz_vector, oy_vector],
                               origin=(oval_new[0], oval_new[1], oval_new[2]), axes=(0, 1, 2), order=1)
        label = fn.affineSlice(self.atlas_label, shape=(oz_length, oy_length), vectors=[oz_vector, oy_vector],
                               origin=(oval_new[0], oval_new[1], oval_new[2]), axes=(0, 1, 2), order=0)
        da_contour = make_contour_img(label)
        self.simg.set_data(atlas, label, da_contour, scale=None)

        self.s_rotm_3d = np.dot(rotation_z(z_angle), rotation_y(-y_angle))

        slide_dist = (self.current_sagital_index - self.Bregma[1])
        ml_plate_verts = self.ml_plate_verts + np.array([slide_dist, 0, 0])
        ml_plate_verts = np.dot(self.s_rotm_3d, (ml_plate_verts - self.rotate_origin_3d).T).T + self.rotate_origin_3d
        ml_plate_md = gl.MeshData(vertexes=ml_plate_verts, faces=self.ml_plate_faces)
        self.ml_plate_mesh.setMeshData(meshdata=ml_plate_md)

    def sagital_slice_page_changed(self, page_number):
        self.clear_all_display_obj()
        if self.atlas_data is None or self.atlas_label is None:
            return
        self.current_sagital_index = page_number
        self.set_perfect_sagittal_slice()
        if self.remain_angle:
            if self.sagittal_rotated:
                self.rotate_sagittal_current_slice()
        else:
            if self.sagittal_rotated:
                self.srotation_ctrl.h_spinbox.blockSignals(True)
                self.srotation_ctrl.v_spinbox.blockSignals(True)
                self.srotation_ctrl.h_slider.setValue(0)
                self.srotation_ctrl.v_slider.setValue(0)
                self.srotation_ctrl.h_spinbox.blockSignals(False)
                self.srotation_ctrl.v_spinbox.blockSignals(False)
                self.sagittal_rotated = False

    def sagital_slice_rotated(self, rads):
        self.clear_all_display_obj()
        if self.atlas_data is None or self.atlas_label is None:
            return
        self.sagittal_rads = rads
        if np.all(np.ravel(rads) == 0):
            if self.sagittal_rotated:
                self.set_perfect_sagittal_slice()
                self.sagittal_rotated = False
        else:
            self.rotate_sagittal_current_slice()
            self.sagittal_rotated = True

    # horizontal page number changed and slice rotated
    def set_perfect_horizontal_slice(self):
        slice_number = self.atlas_size[0] - 1 - self.current_horizontal_index
        da_atlas_slice = self.atlas_data[slice_number, :, :]
        da_atlas_label = self.atlas_label[slice_number, :, :]
        da_atlas_contour = self.atlas_boundary['h_contour'][slice_number, :, :]
        self.himg.set_data(da_atlas_slice, da_atlas_label, da_atlas_contour, scale=None)

        slide_dist = (self.current_horizontal_index - self.origin_3d[2])
        dv_plate_verts = self.dv_plate_verts + np.array([0, 0, slide_dist])
        dv_plate_md = gl.MeshData(vertexes=dv_plate_verts, faces=self.dv_plate_faces)
        self.dv_plate_mesh.setMeshData(meshdata=dv_plate_md)
        self.get_3d_origin()

    def rotate_horizontal_current_slice(self):
        c_id = self.current_coronal_index
        s_id = self.current_sagital_index
        h_id = self.atlas_size[0] - 1 - self.current_horizontal_index

        x_angle = self.horizontal_rads[0]
        y_angle = self.horizontal_rads[1]
        self.h_rotm_2d = np.dot(rotation_z(y_angle), rotation_y(x_angle))

        o_val = np.array([h_id, 0, 0])
        o_rot = np.array([h_id, s_id, c_id])

        ox_vector = np.dot(self.h_rotm_2d, np.array([0, 1, 0]))
        oy_vector = np.dot(self.h_rotm_2d, np.array([0, 0, 1]))

        oval_new = o_rot + np.dot(self.h_rotm_2d, o_val - o_rot)

        oy_length = self.atlas_size[2]
        ox_length = self.atlas_size[1]

        atlas = fn.affineSlice(self.atlas_data, shape=(ox_length, oy_length), vectors=[ox_vector, oy_vector],
                               origin=(oval_new[0], oval_new[1], oval_new[2]), axes=(0, 1, 2), order=1)
        label = fn.affineSlice(self.atlas_label, shape=(ox_length, oy_length), vectors=[ox_vector, oy_vector],
                               origin=(oval_new[0], oval_new[1], oval_new[2]), axes=(0, 1, 2), order=0)
        da_contour = make_contour_img(label)
        self.himg.set_data(atlas, label, da_contour, scale=None)

        self.h_rotm_3d = np.dot(rotation_y(-y_angle), rotation_x(-x_angle))

        slide_dist = (self.Bregma[0] - h_id)
        dv_plate_verts = self.dv_plate_verts + np.array([0, 0, slide_dist])
        dv_plate_verts = np.dot(self.h_rotm_3d, (dv_plate_verts - self.rotate_origin_3d).T).T + self.rotate_origin_3d
        dv_plate_md = gl.MeshData(vertexes=dv_plate_verts, faces=self.dv_plate_faces)
        self.dv_plate_mesh.setMeshData(meshdata=dv_plate_md)

    def horizontal_slice_page_changed(self, page_number):
        self.clear_all_display_obj()
        if self.atlas_data is None or self.atlas_label is None:
            return
        self.current_horizontal_index = page_number

        self.set_perfect_horizontal_slice()
        if self.remain_angle:
            if self.horizontal_rotated:
                self.rotate_horizontal_current_slice()
        else:
            if self.horizontal_rotated:
                self.hrotation_ctrl.h_spinbox.blockSignals(True)
                self.hrotation_ctrl.v_spinbox.blockSignals(True)
                self.hrotation_ctrl.h_slider.setValue(0)
                self.hrotation_ctrl.v_slider.setValue(0)
                self.hrotation_ctrl.h_spinbox.blockSignals(False)
                self.hrotation_ctrl.v_spinbox.blockSignals(False)
                self.horizontal_rotated = False

    def horizontal_slice_rotated(self, rads):
        self.clear_all_display_obj()
        if self.atlas_data is None or self.atlas_label is None:
            return
        self.horizontal_rads = rads
        if np.all(np.ravel(rads) == 0):
            if self.horizontal_rotated:
                self.set_perfect_horizontal_slice()
                self.horizontal_rotated = False
        else:
            self.rotate_horizontal_current_slice()
            self.horizontal_rotated = True

    # get 3d rotation origin
    def get_3d_origin(self):
        c_id = self.current_coronal_index
        s_id = self.current_sagital_index
        h_id = self.current_horizontal_index
        o_rot = np.array([s_id, c_id, h_id])
        self.rotate_origin_3d = o_rot - self.origin_3d

    #
    def rotate_cs_plane_after_merging_probe(self, display_data):
        # display data has to be the merged probe data
        self.cpage_ctrl.set_val(display_data['insertion_vox'][1])
        self.spage_ctrl.set_val(display_data['insertion_vox'][0])
        self.hpage_ctrl.set_val(display_data['insertion_vox'][2])

        ap_sign, ml_sign = get_tilt_sign(display_data['insertion_coords'], display_data['terminus_coords'])

        ap_angle = -1 * ap_sign * display_data['ap_angle']
        ml_angle = ml_sign * display_data['ml_angle']

        start_pnt = display_data['insertion_coords_3d'].copy()
        end_pnt = display_data['terminus_coords_3d'].copy()
        # if ap_angle <= 30 and ap_angle >= -30:
        if ap_angle <= -1e-6 or ap_angle >= 1e-6:
            self.coronal_slice_rotated((0, np.radians(ap_angle)))
            rotated_back_p1 = np.dot(self.c_rotm_3d.T, start_pnt - self.rotate_origin_3d)
            rotated_back_p2 = np.dot(self.c_rotm_3d.T, end_pnt - self.rotate_origin_3d)

            rotated_back_p1 = rotated_back_p1 + self.origin_3d + self.rotate_origin_3d
            rotated_back_p2 = rotated_back_p2 + self.origin_3d + self.rotate_origin_3d

            c_start_pnt = [rotated_back_p1[0], self.atlas_size[0] - rotated_back_p1[2]]
            c_end_pnt = [rotated_back_p2[0], self.atlas_size[0] - rotated_back_p2[2]]
        else:
            c_start_pnt_3d = start_pnt + self.origin_3d
            c_end_pnt_3d = end_pnt + self.origin_3d

            c_start_pnt = [c_start_pnt_3d[0], self.atlas_size[0] - c_start_pnt_3d[2]]
            c_end_pnt = [c_end_pnt_3d[0], self.atlas_size[0] - c_end_pnt_3d[2]]

        if ml_angle <= -1e-6 or ml_angle >= 1e-6:
            self.sagital_slice_rotated((0, np.radians(ml_angle)))
            rotated_back_p1 = np.dot(self.s_rotm_3d.T, start_pnt - self.rotate_origin_3d)
            rotated_back_p2 = np.dot(self.s_rotm_3d.T, end_pnt - self.rotate_origin_3d)

            rotated_back_p1 = rotated_back_p1 + self.origin_3d + self.rotate_origin_3d
            rotated_back_p2 = rotated_back_p2 + self.origin_3d + self.rotate_origin_3d

            s_start_pnt = [rotated_back_p1[1], self.atlas_size[0] - rotated_back_p1[2]]
            s_end_pnt = [rotated_back_p2[1], self.atlas_size[0] - rotated_back_p2[2]]
        else:
            s_start_pnt_3d = start_pnt + self.origin_3d
            s_end_pnt_3d = end_pnt + self.origin_3d

            s_start_pnt = [s_start_pnt_3d[0], self.atlas_size[0] - s_start_pnt_3d[2]]
            s_end_pnt = [s_end_pnt_3d[0], self.atlas_size[0] - s_end_pnt_3d[2]]

        c_pos = np.stack([c_start_pnt, c_end_pnt], axis=0)
        s_pos = np.stack([s_start_pnt, s_end_pnt], axis=0)

        self.cimg.add_display_obj_to_view(c_pos, display_data['vis_color'], 'probe')
        self.simg.add_display_obj_to_view(s_pos, display_data['vis_color'], 'probe')
        self.has_display_objects = True


    def clear_all_display_obj(self):
        if not self.has_display_objects:
            return
        c_n_obj = len(self.cimg.display_objects)
        c_obj_indexes = np.arange(c_n_obj)[::-1]
        for i in c_obj_indexes:
            self.cimg.remove_display_obj_from_view(i)

        s_n_obj = len(self.simg.display_objects)
        s_obj_indexes = np.arange(s_n_obj)[::-1]
        for j in s_obj_indexes:
            self.simg.remove_display_obj_from_view(j)

        self.cimg.display_objects.clear()
        self.simg.display_objects.clear()

        self.has_display_objects = False




    def get_atlas_angles(self):
        c_ang = (self.crotation_ctrl.h_spinbox.value(), self.crotation_ctrl.v_spinbox.value())
        s_ang = (self.srotation_ctrl.h_spinbox.value(), self.srotation_ctrl.v_spinbox.value())
        h_ang = (self.hrotation_ctrl.h_spinbox.value(), self.hrotation_ctrl.v_spinbox.value())
        return [c_ang, s_ang, h_ang]

    def set_volume_atlas_ctrl_data(self, atlas_ctrl_data):
        self.cpage_ctrl.page_slider.setValue(atlas_ctrl_data['coronal_index'])
        self.spage_ctrl.page_slider.setValue(atlas_ctrl_data['sagittal_index'])
        self.hpage_ctrl.page_slider.setValue(atlas_ctrl_data['horizontal_index'])
        atlas_rotation = atlas_ctrl_data['slice_rotation']
        self.crotation_ctrl.h_spinbox.setValue(atlas_rotation[0][0])
        self.crotation_ctrl.v_spinbox.setValue(atlas_rotation[0][1])
        self.srotation_ctrl.h_spinbox.setValue(atlas_rotation[1][0])
        self.srotation_ctrl.v_spinbox.setValue(atlas_rotation[1][1])
        self.hrotation_ctrl.h_spinbox.setValue(atlas_rotation[2][0])
        self.hrotation_ctrl.v_spinbox.setValue(atlas_rotation[2][1])

    def get_coronal_3d(self, points2, coronal_index=None):
        if coronal_index is None:
            da_y = np.ones(len(points2)) * self.current_coronal_index
        else:
            da_y = coronal_index
        points3 = np.vstack([points2[:, 0], da_y, self.atlas_size[0] - points2[:, 1]]).T
        points3 = points3 - self.origin_3d
        if self.coronal_rotated:
            rot_mat = self.c_rotm_3d
            rotation_origin = self.rotate_origin_3d
            points3 = np.dot(rot_mat, (points3 - rotation_origin).T).T + rotation_origin
        return points3

    def get_sagittal_3d(self, points2, sagital_index=None):
        if sagital_index is None:
            da_x = np.ones(len(points2)) * self.current_sagital_index
        else:
            da_x = sagital_index
        points3 = np.vstack([da_x, points2[:, 0], self.atlas_size[0] - points2[:, 1]]).T
        points3 = points3 - self.origin_3d
        if self.sagittal_rotated:
            rot_mat = self.s_rotm_3d
            rotation_origin = self.rotate_origin_3d
            points3 = np.dot(rot_mat, (points3 - rotation_origin).T).T + rotation_origin
        return points3

    def get_horizontal_3d(self, points2, horizontal_index=None):
        if horizontal_index is None:
            da_z = np.ones(len(points2)) * self.current_horizontal_index
        else:
            da_z = horizontal_index
        points3 = np.vstack([points2[:, 1], points2[:, 0], da_z]).T
        points3 = points3 - self.origin_3d
        if self.horizontal_rotated:
            rot_mat = self.h_rotm_3d
            rotation_origin = self.rotate_origin_3d
            points3 = np.dot(rot_mat, (points3 - rotation_origin).T).T + rotation_origin
        return points3

    def get_3d_data_from_2d_view(self, processing_data, atlas_display):
        if atlas_display == 'coronal':
            data = self.get_coronal_3d(processing_data)
        elif atlas_display == 'sagittal':
            data = self.get_sagittal_3d(processing_data)
        else:
            data = self.get_horizontal_3d(processing_data)
        return data

    def pre_trajectory_changed(self):
        self.working_atlas.remove_pre_trajectories_vis_lines()

    def get_pre_vis_data_for_volume_atlas(self, data, base_loc):
        base_loc = np.ravel(base_loc) / self.vox_size_um
        start_pnt, end_pnt = rotate_base_points(data, base_loc)
        return start_pnt, end_pnt

    def get_pre_vis_data_for_slice_atlas(self, data, base_loc):
        start_pnt, end_pnt = rotate_base_points(data, base_loc)
        return start_pnt, end_pnt

    def set_pre_vis_data(self, start_pnt, end_pnt):
        line_data = []
        for i in range(len(start_pnt)):
            da_pnts = np.asarray([start_pnt[i], end_pnt[i]])
            line_data.append(da_pnts)
        self.working_atlas.set_pre_design_vis_data(line_data)
        self.working_atlas.image_dict['atlas-probe'].setData(pos=np.asarray(start_pnt))

    def draw_pre_2d_vis_data_for_volume_atlas(self, data, base_loc):
        if len(data) == 1:
            base_loc = np.ravel(base_loc) / self.vox_size_um
            temp = np.zeros((len(base_loc), 2))
            temp[:, 0] = base_loc + data[0][0]
            temp[:, 1] = data[0][1]
            self.working_atlas.image_dict['atlas-probe'].setData(pos=np.asarray(temp))
        elif len(data) == 2:
            start_pnt, end_pnt = self.get_pre_vis_data_for_volume_atlas(data, base_loc)
            temp = np.vstack([start_pnt, end_pnt])
            self.working_atlas.image_dict['atlas-probe'].setData(pos=temp)
        else:
            self.working_atlas.image_dict['atlas-probe'].clear()
            self.working_atlas.remove_pre_trajectories_vis_lines()

    def draw_pre_2d_vis_data_for_slice_atlas(self, data, base_loc):
        if self.slice_info_ready:
            base_loc = np.ravel(base_loc) / 1000 * self.slice_size[1] / self.slice_width
            if len(data) == 1:
                temp = np.zeros((len(base_loc), 2))
                temp[:, 0] = base_loc + data[0][0]
                temp[:, 1] = data[0][1]
                self.working_atlas.image_dict['atlas-probe'].setData(pos=np.asarray(temp))
            elif len(data) == 2:
                start_pnt, end_pnt = self.get_pre_vis_data_for_slice_atlas(data, base_loc)
                temp = np.vstack([start_pnt, end_pnt])
                self.working_atlas.image_dict['atlas-probe'].setData(pos=temp)
            else:
                self.working_atlas.image_dict['atlas-probe'].clear()
                self.working_atlas.remove_pre_trajectories_vis_lines()
        else:
            if len(data) == 1:
                self.working_atlas.image_dict['atlas-probe'].setData(pos=np.asarray(data))
            elif len(data) == 2:
                self.working_atlas.image_dict['atlas-probe'].setData(pos=np.asarray(data))
                self.working_atlas.set_pre_design_vis_data([np.asarray(data)])
            else:
                self.working_atlas.image_dict['atlas-probe'].clear()
                self.working_atlas.remove_pre_trajectories_vis_lines()


    # def get_pre_multi_probe_data(self, data, atlas_display, multi_settings):
    #     x_vals = multi_settings['x_vals']
    #     y_vals = multi_settings['y_vals']
    #     n_probes = len(x_vals)
    #
    #     if atlas_display == 'coronal':
    #         data3 = self.get_coronal_3d(data)
    #     elif atlas_display == 'sagittal':
    #         data3 = self.get_sagittal_3d(data)
    #     else:
    #         data3 = self.get_horizontal_3d(data)
    #
    #     pdata = np.asarray(data3)
    #     p_order = np.argsort(pdata[:, 1])
    #     pdata = pdata[p_order, :]
    #
    #     direction = pdata[0] - pdata[1]
    #     direction = direction / np.linalg.norm(direction)
    #
    #     rot_m = get_direction_rotation(direction)
    #     base_end = np.dot(rot_m.T, pdata[1] - pdata[0])
    #
    #     base_start_3d = np.zeros((n_probes, 3))
    #     base_start_3d[:, 1] = x_vals.copy()
    #     base_start_3d[:, 2] = y_vals.copy()
    #
    #     base_end_3d = base_start_3d.copy()
    #     base_end_3d[:, 0] = base_end
    #
    #     points3_list = []
    #     for i in range(n_probes):
    #         tp3 = np.vstack([base_start_3d[i], base_end_3d[i]])
    #         points3 = np.dot(rot_m, tp3.T).T
    #         points3_list.append(points3)
    #
    #     return points3_list

    def get_multi_probe_2d_vis_data(self, data, multi_settings):
        x_vals = multi_settings['x_vals']
        y_vals = multi_settings['y_vals']
        n_probes = len(x_vals)

        vind = np.where(y_vals == 0)[0]
        if len(vind) == 0:
            vis_data = [data]
        else:
            vis_data = []
            base_loc = np.ravel(x_vals)[vind]

    def get_plane_norm_vector(self, atlas_display):
        if atlas_display == 'coronal':
            if self.coronal_rotated:
                n_vec = np.dot(self.c_rotm_3d, np.array([0, -1, 0]))
            else:
                n_vec = np.array([0, -1, 0])
        elif atlas_display == 'sagittal':
            if self.sagittal_rotated:
                n_vec = np.dot(self.s_rotm_3d, np.array([1, 0, 0]))
            else:
                n_vec = np.array([1, 0, 0])
        else:
            if self.horizontal_rotated:
                n_vec = np.dot(self.h_rotm_3d, np.array([0, 0, 1]))
            else:
                n_vec = np.array([0, 0, 1])
        return n_vec


    #
    def draw_volume_pre_trajectory(self, data, n_pre_trajectory):
        if len(data) == 1:
            if n_pre_trajectory == 1:

                self.working_atlas.image_dict['atlas-probe'].setData(pos=np.asarray(data))
            else:
                base_loc = np.array([-375, -125, 125, 375]) / self.vox_size_um
                temp = np.stack([data[0][0] + base_loc, np.repeat(data[0][1], 4)], axis=1)
                self.working_atlas.image_dict['atlas-probe'].setData(pos=np.asarray(temp))
        elif len(data) == 2:
            if n_pre_trajectory == 1:
                print('data', data)
                start_pnt, end_pnt = rotate_base_points(np.asarray(data), np.array([0]))
                print(start_pnt, end_pnt)
                self.working_atlas.image_dict['atlas-probe'].setData(pos=np.asarray(data))
            else:
                base_loc = np.array([-375, -125, 125, 375]) / self.vox_size_um
                start_pnt, end_pnt = rotate_base_points(np.asarray(data), base_loc)
                temp = np.vstack([start_pnt, end_pnt])
                self.working_atlas.image_dict['atlas-probe'].setData(pos=temp)
        else:
            self.working_atlas.image_dict['atlas-probe'].clear()

    def vis_3d_planes(self, vis):
        self.ap_plate_mesh.setVisible(vis)
        self.dv_plate_mesh.setVisible(vis)
        self.ml_plate_mesh.setVisible(vis)

    def vis_3d_axes(self, vis):
        self.axis.setVisible(vis)

    # ----------- clear atlas
    def clear_atlas_stacks(self):
        valid_keys = ['atlas-overlay', 'atlas-contour', 'atlas-mask', 'circle_follow', 'lasso_path', 'atlas-virus',
                      'atlas-probe', 'atlas-cells', 'atlas-drawing']
        for da_key in valid_keys:
            self.working_atlas.image_dict[da_key].clear()

    def clear_volume_atlas(self):
        if self.atlas_label is None or self.atlas_data is None:
            return
        valid_keys = ['atlas-overlay', 'atlas-contour', 'atlas-mask', 'circle_follow', 'lasso_path', 'atlas-virus',
                      'atlas-probe', 'atlas-cells', 'atlas-drawing']
        for da_key in valid_keys:
            self.cimg.image_dict[da_key].clear()
            self.simg.image_dict[da_key].clear()
            self.himg.image_dict[da_key].clear()
        self.cimg.img.clear()
        self.simg.img.clear()
        self.himg.img.clear()


    def clear_slice_atlas(self):
        pass

    def clear_atlas(self):
        self.clear_volume_atlas()
        self.clear_slice_atlas()















import urllib.request
import numpy as np
import time
import os
import sys
from os.path import dirname, realpath, join
from PyQt5.QtWidgets import *
from PyQt5.QtGui import *
from PyQt5.QtCore import *

import pickle
import shutil
import requests

from .atlas_loader import process_atlas_raw_data
from .obj_items import render_volume, render_small_volume


class DownloadThread(QThread):
    download_process_signal = pyqtSignal(int)

    def __init__(self, url, filesize, fileobj, buffer):
        super(DownloadThread, self).__init__()
        self.url = url
        self.filesize = filesize
        self.fileobj = fileobj
        self.buffer = buffer

    def run(self):
        try:
            rsp = requests.get(self.url, stream=True)
            offset = 0
            for chunk in rsp.iter_content(chunk_size=self.buffer):
                if not chunk:
                    break
                self.fileobj.seek(offset)
                self.fileobj.write(chunk)
                offset = offset + len(chunk)
                proess = offset / int(self.filesize) * 100
                self.download_process_signal.emit(int(proess))
            self.fileobj.close()
            self.exit(0)

        except Exception as e:
            print(e)


class WorkerProcessData(QObject):
    finished = pyqtSignal()
    progress = pyqtSignal(int)

    def __init__(self):
        super(WorkerProcessData, self).__init__()
        self.saving_folder = None
        self.data_local = None
        self.segmentation_local = None
        self.mask_local = None
        self.b_val = None
        self.l_val = None
        self.vox_size = None

        self.atlas_data = None
        self.atlas_info = None
        self.segmentation_data = None
        self.unique_label = None
        self.label_info = None
        self.boundary = None
        self.small_mesh_list = None
        self.mesh_data = None

    def set_data(self, saving_folder, data_local, segmentation_local, mask_local, b_val, l_val, vox_size):
        self.saving_folder = saving_folder
        self.data_local = data_local
        self.segmentation_local = segmentation_local
        self.mask_local = mask_local
        self.b_val = b_val
        self.l_val = l_val
        self.vox_size = vox_size

    def run(self):
        target = os.path.join(self.saving_folder, 'atlas_labels.pkl')
        if not os.path.exists(target):
            shutil.copyfile(join(dirname(__file__), "data/atlas_labels.pkl"), target)

        self.progress.emit(1)

        infile = open(os.path.join(self.saving_folder, 'atlas_labels.pkl'), 'rb')
        self.label_info = pickle.load(infile)
        infile.close()

        self.progress.emit(5)

        self.atlas_data, self.atlas_info, self.segmentation_data, self.unique_label, self.boundary, msg = \
            process_atlas_raw_data(self.saving_folder, data_file=self.data_local,
                                   segmentation_file=self.segmentation_local, mask_file=self.mask_local,
                                   bregma_coordinates=self.b_val, lambda_coordinates=self.l_val,
                                   voxel_size=self.vox_size)

        if msg == 'Atlas loaded successfully.':
            self.progress.emit(40)
            self.mesh_data = render_volume(self.atlas_data, self.saving_folder, factor=2, level=0.1)
            self.progress.emit(50)
            save_path = os.path.join(self.saving_folder, 'meshes')
            if not os.path.exists(save_path):
                os.mkdir(save_path)

            n_label = len(self.unique_label)
            progress_step = np.linspace(50, 95, n_label)

            for i in range(n_label):
                self.progress.emit(progress_step[i])
                label_id = int(self.unique_label[i])

                if label_id == 0:
                    continue
                render_small_volume(label_id, save_path, self.atlas_data, self.segmentation_data, factor=2, level=0.1)

            self.small_mesh_list = {}
            file_list = os.listdir(save_path)
            for da_file in file_list:
                file_name = os.path.basename(da_file)
                da_name, file_extension = os.path.splitext(file_name)
                if file_extension == '.pkl':
                    infile = open(os.path.join(save_path, da_file), 'rb')
                    md = pickle.load(infile)
                    infile.close()

                    self.small_mesh_list[str(da_name)] = md

            self.progress.emit(97)
            outfile = open(os.path.join(self.saving_folder, 'atlas_small_meshdata.pkl'), 'wb')
            pickle.dump(self.small_mesh_list, outfile)
            outfile.close()

            self.progress.emit(100)

        self.finished.emit()


class AtlasDownloader(QDialog):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        layout = QVBoxLayout(self)

        self.setWindowTitle("Waxholm Rat Atlas Downloader")

        self.continue_process = True

        self.label_url = "https://www.nitrc.org/frs/download.php/12261/WHS_SD_rat_atlas_v4.label"
        self.data_url = "https://www.nitrc.org/frs/downloadlink.php/9423/WHS_SD_rat_T2star_v1.01.nii.gz"
        self.mask_url = "https://www.nitrc.org/frs/downloadlink.php/9425/WHS_SD_rat_brainmask_v1.01.nii.gz"
        self.segmentation_url = "https://www.nitrc.org/frs/download.php/12260/WHS_SD_rat_atlas_v4.nii.gz"

        self.label_local = "WHS_SD_rat_atlas_v4.label"
        self.data_local = "WHS_SD_rat_T2star_v1.01.nii.gz"
        self.mask_local = "WHS_SD_rat_brainmask_v1.01.nii.gz"
        self.segmentation_local = "WHS_SD_rat_atlas_v4.nii.gz"

        self.saving_folder = None

        self.thread = QThread()
        self.worker = WorkerProcessData()

        self.b_val = (246, 653, 440)
        self.l_val = (244, 442, 464)
        self.vox_size = 39.0625

        self.finish = [False, False, False, False]
        self.process_finished = False

        self.label_bar = QProgressBar()
        self.label_bar.setMinimumWidth(400)
        self.label_bar.setValue(0)

        self.data_bar = QProgressBar()
        self.data_bar.setMinimumWidth(400)
        self.data_bar.setValue(0)

        self.mask_bar = QProgressBar()
        self.mask_bar.setMinimumWidth(400)
        self.mask_bar.setValue(0)

        self.segmentation_bar = QProgressBar()
        self.segmentation_bar.setMinimumWidth(400)
        self.segmentation_bar.setValue(0)

        self.download_btn = QPushButton()
        self.download_btn. setMinimumWidth(100)
        self.download_btn.setText("Download")

        self.process_btn = QPushButton()
        self.process_btn.setMinimumWidth(100)
        self.process_btn.setText("Process")

        self.process_label = QLabel()
        self.process_label.setMinimumWidth(100)
        self.process_label.setText("")

        self.progress = QProgressBar(self)
        self.progress.setMinimumWidth(100)

        self.process_info = QLabel('The whole process takes around 40 min - 1 hour. \n'
                                   'This window will be closed automatically when processing finished.\n')

        # ok button, used to close window
        ok_btn = QDialogButtonBox(QDialogButtonBox.Ok)
        ok_btn.accepted.connect(self.accept)

        layout.addWidget(self.label_bar)
        layout.addWidget(self.data_bar)
        layout.addWidget(self.mask_bar)
        layout.addWidget(self.segmentation_bar)
        layout.addWidget(self.download_btn)
        layout.addWidget(self.process_info)
        layout.addWidget(self.process_btn)
        layout.addWidget(self.progress)
        layout.addWidget(self.process_label)
        # layout.addWidget(ok_btn)

        # Binding Button Event
        self.download_btn.clicked.connect(self.download_start)
        self.process_btn.clicked.connect(self.process_start)

    # Download button event
    def download_start(self):
        self.saving_folder = str(QFileDialog.getExistingDirectory(self, "Select Folder to Save Atlas"))

        if self.saving_folder != '':
            self.download_btn.setVisible(False)
            target = os.path.join(self.saving_folder, 'atlas_labels.pkl')
            if not os.path.exists(target):
                shutil.copyfile(join(dirname(__file__), "data/atlas_labels.pkl"), target)

            self.start_thread(self.label_url, self.label_local, self.set_label_bar_value)
            self.start_thread(self.mask_url, self.mask_local, self.set_mask_bar_value)
            time.sleep(0.1)
            self.start_thread(self.segmentation_url, self.segmentation_local, self.set_segmentation_bar_value)
            time.sleep(0.1)
            self.start_thread(self.data_url, self.data_local, self.set_data_bar_value)

    #
    def start_thread(self, url, local, func):
        file_size = requests.get(url, stream=True).headers['Content-Length']
        file_obj = open(os.path.join(self.saving_folder, local), 'wb')

        self.da_thread = DownloadThread(url, file_size, file_obj, buffer=1024)
        self.da_thread.download_process_signal.connect(func)
        self.da_thread.start()

    # Setting progress bar
    def set_label_bar_value(self, value):
        self.label_bar.setValue(value)
        if value == 100:
            self.finish[0] = True
            return

    def set_data_bar_value(self, value):
        self.data_bar.setValue(value)
        if value == 100:
            self.finish[1] = True
            return

    def set_mask_bar_value(self, value):
        self.mask_bar.setValue(value)
        if value == 100:
            self.finish[2] = True
            return

    def set_segmentation_bar_value(self, value):
        self.segmentation_bar.setValue(value)
        if value == 100:
            self.finish[3] = True
            return

    def report_progress(self, i):
        self.progress.setValue(i)

    def process_start(self):
        # if not np.all(self.finish):
        #     return
        if self.saving_folder is not None:
            saving_folder = self.saving_folder
        else:
            saving_folder = str(QFileDialog.getExistingDirectory(self, "Select Atlas Folder"))

        if saving_folder != '':
            self.process_btn.setVisible(False)
            self.worker.set_data(saving_folder, self.data_local, self.segmentation_local, self.mask_local,
                                 self.b_val, self.l_val, self.vox_size)
            self.worker.moveToThread(self.thread)
            self.thread.started.connect(self.worker.run)
            self.worker.finished.connect(self.on_finish)
            # self.worker.finished.connect(self.worker.deleteLater)
            self.thread.finished.connect(self.thread.deleteLater)
            self.worker.progress.connect(self.report_progress)
            self.thread.start()

    def on_finish(self):
        self.thread.quit()
        self.close()

    def closeEvent(self, event):
        reply = QMessageBox.question(self, 'Message',
                                     "Do you want to leave?", QMessageBox.Yes, QMessageBox.No)

        if reply == QMessageBox.Yes:
            self.continue_process = False
            event.accept()
        else:
            event.ignore()

    # def accept(self) -> None:
    #     if self.process_finished:
    #         self.close()
    #     else:
    #         return





import os
import sys
import numpy as np
import pyqtgraph as pg
from PyQt5.QtCore import *
from PyQt5.QtGui import *

from .movable_points import TriangulationPoints


class ClickableSlice(pg.ImageItem):
    class SignalProxy(QObject):
        mouseHovered = pyqtSignal(object)  # id
        mouseClicked = pyqtSignal(object)  # id

    def __init__(self):
        self._sigprox = ClickableSlice.SignalProxy()
        self.mouseHovered = self._sigprox.mouseHovered
        self.mouseClicked = self._sigprox.mouseClicked

        pg.ImageItem.__init__(self)

        self.setAcceptHoverEvents(True)
        self.setLevels((0, 1))
        self.setOpts(axisOrder='row-major')

    def set_data(self, atlas_data, scale=None):
        # self.label_data = label
        # self.atlas_data = atlas_data
        if scale is not None:
            self.resetTransform()
            self.scale(*scale)
        self.setImage(atlas_data, autoLevels=False)
        # self.label_img.setImage(self.label_data, autoLevels=False)

    def hoverEvent(self, event):
        if event.isExit():
            return
        try:
            pos = (event.pos())
        except (IndexError, AttributeError):
            return
        self.mouseHovered.emit(event)

    def mouseClickEvent(self, event):
        pos = (event.pos().x(), event.pos().y())
        # id = self.label_data[int(event.pos().x()), int(event.pos().y())]
        self.mouseClicked.emit(pos)
        # self.mouseClicked.emit([event, id])


class SliceStacks(pg.GraphicsLayoutWidget):
    class SignalProxy(QObject):
        mouseHovered = pyqtSignal(object)  # id
        mouseClicked = pyqtSignal(object)  # id

    def __init__(self):
        self._sigprox = SliceStacks.SignalProxy()
        self.sig_mouse_hovered = self._sigprox.mouseHovered
        self.sig_mouse_clicked = self._sigprox.mouseClicked

        pg.GraphicsLayoutWidget.__init__(self)
        self.vb = self.addViewBox()
        self.setBackground('k')
        self.vb.setAspectLocked()
        self.vb.invertY(True)
        self.vb.setAcceptHoverEvents(True)

        # self.img = SliceStack()
        self.img = ClickableSlice()
        self.img.mouseClicked.connect(self.mouse_clicked)
        self.img.mouseHovered.connect(self.mouse_hovered)
        self.label_img = pg.ImageItem()
        self.label_img.setCompositionMode(QPainter.CompositionMode.CompositionMode_Multiply)
        self.boundary = pg.ImageItem()
        self.boundary.setCompositionMode(QPainter.CompositionMode.CompositionMode_Plus)
        self.boundary.setVisible(False)

        self.display_objects = []

        tri_pnts = TriangulationPoints()
        self.tri_lines_list = []

        circle_follow = pg.PlotDataItem(pen=pg.mkPen('r', width=2, style=Qt.DashLine))
        lasso_path = pg.PlotDataItem(pen=pg.mkPen(color=(0, 255, 255), width=3, style=Qt.DashLine),
                                     symbolPen=(0, 255, 255), symbol='o', symbolSize=4)
        ruler_path = pg.PlotDataItem(pen=pg.mkPen(color='y', width=3, style=Qt.DashLine),
                                     symbolPen='y', symbolBrush='y', symbol='s', symbolSize=3)

        overlay_img = pg.ImageItem()

        overlay_contour = pg.ImageItem()
        overlay_contour.setLevels(levels=(0, 1))

        mask_img = pg.ImageItem()
        mask_img.setLevels(levels=(0, 1))

        grid_lines = pg.GridItem(pen=(128, 128, 128))
        grid_lines.setVisible(False)

        virus_pnts = pg.ScatterPlotItem(pen=(133, 255, 117), brush=(133, 255, 117), symbol='s', size=5, hoverSize=8)
        probe_pnts = pg.ScatterPlotItem(pen=(0, 0, 255), brush=(0, 0, 255), symbol='s', size=5, hoverSize=8)
        cell_pnts = pg.ScatterPlotItem(pen=(0, 255, 0), brush=(0, 255, 0), size=5, hoverSize=8)
        drawing_pnts = pg.PlotDataItem(pen=pg.mkPen(color=(255, 102, 0), width=2), brush=None)
        contour_pnts = pg.PlotDataItem(pen=pg.mkPen(color=(255, 0, 255), width=3), brush=None)
        probe_trajectory = pg.PlotDataItem(pen=pg.mkPen(color=(0, 0, 255), width=2), brush=None)

        self.pre_trajectory_list = [pg.PlotDataItem(pen=pg.mkPen(color=(0, 0, 255), width=2), brush=None)]

        self.image_dict = {'atlas-overlay': overlay_img,
                           'atlas-mask': mask_img,
                           'grid_lines': grid_lines,
                           'tri_pnts': tri_pnts,
                           'circle_follow': circle_follow,
                           'lasso_path': lasso_path,
                           'atlas-virus': virus_pnts,
                           'atlas-contour': contour_pnts,
                           'atlas-probe': probe_pnts,
                           'atlas-cells': cell_pnts,
                           'ruler_path': ruler_path,
                           'atlas-drawing': drawing_pnts,
                           'atlas-trajectory': probe_trajectory}
        self.image_dict_keys = list(self.image_dict.keys())

        self.v_line = pg.InfiniteLine(angle=90, movable=False)
        self.h_line = pg.InfiniteLine(angle=0, movable=False)
        self.v_line.setVisible(False)
        self.h_line.setVisible(False)

        self.vb.addItem(self.img)
        self.vb.addItem(self.label_img)

        for i in range(len(self.image_dict)):
            self.vb.addItem(self.image_dict[self.image_dict_keys[i]])

        self.vb.addItem(self.pre_trajectory_list[0])

        self.vb.addItem(self.boundary)

        # self.vb.addItem(self.after_trajectory)

        self.vb.addItem(self.v_line, ignoreBounds=True)
        self.vb.addItem(self.h_line, ignoreBounds=True)

    def set_pre_design_vis_data(self, data):
        n_probe = len(data)
        exist_n_probes = len(self.pre_trajectory_list)
        if n_probe < exist_n_probes:
            del_ind = np.arange(n_probe, exist_n_probes)[::-1]
            for i in del_ind:
                self.vb.removeItem(self.pre_trajectory_list[i])
                self.pre_trajectory_list[i].deleteLater()
                del self.pre_trajectory_list[i]
        elif n_probe > exist_n_probes:
            for i in range(exist_n_probes, n_probe):
                self.pre_trajectory_list.append(pg.PlotDataItem(pen=pg.mkPen(color=(0, 0, 255), width=2), brush=None))
                self.vb.addItem(self.pre_trajectory_list[-1])

        for i in range(n_probe):
            self.pre_trajectory_list[i].setData(data[i])

    def pre_trajectory_color_changed(self, col, lw):
        for i in range(len(self.pre_trajectory_list)):
            self.pre_trajectory_list[i].setPen(pg.mkPen(color=col, width=lw))

    def remove_pre_trajectories_vis_lines(self):
        del_ind = np.arange(len(self.pre_trajectory_list))[::-1]
        for i in del_ind:
            self.vb.removeItem(self.pre_trajectory_list[i])
            self.pre_trajectory_list[i].deleteLater()
            del self.pre_trajectory_list[i]
            # self.pre_trajectory_list[i].updateItems(True)
        self.pre_trajectory_list.clear()


    def set_data(self, atlas, label, contour, scale=None):
        self.img.set_data(atlas)
        self.label_img.setImage(label, autoLevels=False)
        self.boundary.setImage(contour, atutoLevels=False)

    def add_display_obj_to_view(self, obj_data, obj_color, obj_type):
        if obj_type == 'probe':
            obj_vis = pg.PlotDataItem(pen=pg.mkPen(color=obj_color, width=2),
                                      symbolPen=obj_color, symbolBrush=None, symbol='s', symbolSize=3)
            obj_vis.setData(np.asarray(obj_data))
            self.display_objects.append(obj_vis)
            self.vb.addItem(self.display_objects[-1])

    def remove_display_obj_from_view(self, obj_index):
        self.vb.removeItem(self.display_objects[obj_index])
        self.display_objects[obj_index].deleteLater()
        del self.display_objects[obj_index]

    def boundingRect(self):
        return self.label_img.boundingRect()

    def set_lut(self, lut):
        self.label_img.setLookupTable(lut)

    def set_overlay(self, overlay):
        mode = getattr(QPainter, 'CompositionMode_' + overlay)
        self.label_img.setCompositionMode(mode)

    def set_label_opacity(self, o):
        self.label_img.setOpacity(o)

    # def set_label_colors(self, colors):
    #     self.label_colors = colors

    def mouse_hovered(self, event):
        if event.isExit():
            return
        try:
            pos = (event.pos())
        except (IndexError, AttributeError):
            return
        self.sig_mouse_hovered.emit(pos)

    def mouse_clicked(self, pos):
        self.sig_mouse_clicked.emit(pos)

